
#------------------------------------------------------
# VERY UNSTABLE API TO EXECUTE FRONTEND PROGRAMS USING V0 EXECUTORS
#------------------------------------------------------
from functools import lru_cache
import json
from v0.relationalai import debugging
from v0.relationalai.semantics.lqp.executor import LQPExecutor
from v0.relationalai.semantics.rel.executor import RelExecutor
from v0.relationalai.semantics.metamodel import ir as v0, factory as v0_factory
from v0.relationalai.semantics.metamodel.visitor import collect_by_type
from v0.relationalai.semantics.snowflake import Table as v0Table
try:
    from v0.relationalai.clients.snowflake import Provider as v0Provider #type: ignore
except ImportError:
    from v0.relationalai.clients.resources.snowflake import Provider as v0Provider
from v0.relationalai.clients.config import Config

# from ..config import Config
from ..config.shims import DRY_RUN
from ..semantics import Model, Fragment
from ..semantics.metamodel import metamodel as mm
from ..semantics.metamodel.typer import Typer
from ..semantics.metamodel.metamodel_analyzer import Normalize
from .mm2v0 import Translator

DEBUG=False
PRINT_RESULT=False
TYPER_DEBUGGER=False

# DEBUG=True
# PRINT_RESULT=True
# TYPER_DEBUGGER=True

@lru_cache()
def get_config():
    return Config()

def with_source(item: mm.Node):
    if not hasattr(item, "source"):
        raise ValueError(f"Item {item} has no source")
    elif item.source is None:
        return {}
    elif debugging.DEBUG:
        source = item.source.block
        if source:
            return { "file": source.file, "line": source.line, "source": source.source }
        else:
            return {"file":item.source.file, "line":item.source.line}
    else:
        return {"file":item.source.file, "line":item.source.line}

def execute(query: Fragment, model: Model|None = None, executor=None, export_to="", update=False):
    if not executor:
        # use_lqp = Config().reasoner.rule.use_lqp
        use_lqp = bool(get_config().get("reasoner.rule.use_lqp", True))
        executor = "lqp" if use_lqp else "rel"
    mm_model = model.to_metamodel() if model else None
    mm_query = query.to_metamodel()
    assert isinstance(mm_query, mm.Node)
    return execute_mm(mm_query, mm_model, executor, export_to=export_to, update=update, model=model)

def execute_mm(mm_query: mm.Task, mm_model: mm.Model|None = None, executor="lqp", export_to="", update=False, model: Model|None = None):
    # perform type inference
    typer = Typer(enforce=False)
    # normalize the metamodel
    normalizer = Normalize()
    # translate the metamodel into a v0 query
    translator = Translator()

    # for typer debugging
    debugger_msgs = []

    try:
        #------------------------------------------------------
        # Model processing
        #------------------------------------------------------
        v0_model = None
        if mm_model:
            # type inference
            debugger_msgs.append(json.dumps({ 'id': 'model', 'content': str(mm_model)}))
            with debugging.span("compile", metamodel=mm_model) as span:
                span["compile_type"] = "model_v1"
                mm_model = typer.infer_model(mm_model)
                span["typed_mm"] = str(mm_model)
                debugger_msgs.append(json.dumps({ 'id': 'typed_model', 'content': str(mm_model)}))
                assert(typer.model_net is not None)
                debugger_msgs.append(json.dumps({ 'id': 'model_net', 'content': str(typer.model_net.to_mermaid())}))
                # normalization
                mm_model = mm_model.mut(root=normalizer.normalize(mm_model.root)) # type: ignore
                assert isinstance(mm_model, mm.Model)
                if DEBUG:
                    print("V1 Model:")
                    print(mm_model)
                # translation
                v0_model = translator.translate_model(mm_model)
                if DEBUG:
                    print("Translated v0 Model:")
                    print(v0_model)

        #------------------------------------------------------
        # Query processing
        #------------------------------------------------------
        # type inference
        debugger_msgs.append(json.dumps({ 'id': 'query', 'content': str(mm_query)}))
        with debugging.span("compile", metamodel=mm_query) as span:
            span["compile_type"] = "query_v1"
            mm_query = typer.infer_query(mm_query)
            span["typed_mm"] = str(mm_query)
            debugger_msgs.append(json.dumps({ 'id': 'typed_query', 'content': str(mm_query)}))
            assert(typer.last_net is not None)
            debugger_msgs.append(json.dumps({ 'id': 'query_net', 'content': str(typer.last_net.to_mermaid())}))
            # normalization
            mm_query = normalizer.normalize(mm_query)  # type: ignore
            assert isinstance(mm_query, mm.Task)
            if DEBUG:
                print("V1 Query:")
                print(mm_query)
            # translation
            v0_query = translator.translate_query(mm_query)
            if DEBUG:
                print("Translated v0 Query:")
                print(v0_query)
            assert isinstance(v0_query, v0.Task)

        if v0_model is None:
            # there was no model, so create one from the elements refered to by the query
            v0_model = v0_factory.model(
                collect_by_type(v0.Engine, v0_query),
                collect_by_type(v0.Relation, v0_query),
                v0_factory._collect_reachable_types(collect_by_type(v0.Type, v0_query)),
                v0_factory.logical([])
            )
    finally:
        if TYPER_DEBUGGER:
            with open("typer_debug.jsonl", "w") as f:
                for msg in debugger_msgs:
                    f.write(msg)
                    f.write('\n')

    if DRY_RUN or get_config().get("compiler.dry_run", False):
        results = []
    else:
        # create snowflake tables for all the tables that have been used
        ts = [v0Table(t.name) for t in translator.used_tables if not t.uri.startswith("dataframe://")]
        for t in ts:
            t._lazy_init()
            v0Table._used_sources.add(t)

        export_table = None
        if export_to:
            export_table = v0Table(export_to)

        # get an executor and execute
        executor = _get_executor(executor, model.name if model else "")
        with debugging.span("query", tag=None, export_to=export_to, dsl="", **with_source(mm_query)) as query_span:
            if isinstance(executor, (LQPExecutor, RelExecutor)):
                results = executor.execute(v0_model, v0_query, export_to=export_table, update=update)
            else:
                results = executor.execute(v0_model, v0_query)
            query_span["results"] = results
        if DEBUG or PRINT_RESULT:
            print(results)
    return results

@lru_cache()
def _get_executor(name: str, database: str = "ttb_test"):
    if name == "duckdb":
        from v0.relationalai.semantics.sql.executor.duck_db import DuckDBExecutor
        return DuckDBExecutor()
    elif name == "lqp":
        return LQPExecutor(database)
    elif name == "rel":
        return RelExecutor(database)
    elif name == "snowflake":
        from v0.relationalai.semantics.sql.executor.snowflake import SnowflakeExecutor
        return SnowflakeExecutor(database="TEST_DB", schema="PUBLIC")
    else:
        raise ValueError(f"Unknown executor: {name}")

def get_provider():
    return v0Provider()



from __future__ import annotations

from relationalai.semantics.std import floats

from . import StringValue, IntegerValue, DateValue, DateTimeValue, math, common
from ..frontend.base import Aggregate, Library, Concept, MetaRef, NumberConcept, Expression, Field, Literal, Variable
from ..frontend.core import Float, Number, String, Integer, Date, DateTime
from ..frontend import core
from .. import select

from typing import Union, Literal
import datetime as dt

# the front-end library object
library = Library("datetime")


#--------------------------------------------------
# Format String Constants
#--------------------------------------------------

class ISO:
    DATE = "yyyy-mm-dd"
    HOURS = "yyyy-mm-ddTHH"
    HOURS_TZ = "yyyy-mm-ddTHHz"
    MINUTES = "yyyy-mm-ddTHH:MM"
    MINUTES_TZ = "yyyy-mm-ddTHH:MMz"
    SECONDS = "yyyy-mm-ddTHH:MM:SS"
    SECONDS_TZ = "yyyy-mm-ddTHH:MM:SSz"
    MILLIS = "yyyy-mm-ddTHH:MM:SS.s"
    MILLIS_TZ = "yyyy-mm-ddTHH:MM:SS.sz"



#--------------------------------------------------
# Date
#--------------------------------------------------

# Constructors
_construct_date = library.Relation("construct_date", [Field.input("year", Integer), Field.input("month", Integer), Field.input("day", Integer), Field("date", Date)])
_construct_date_from_datetime = library.Relation("construct_date_from_datetime", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("date", Date)])
_construct_datetime_ms_tz = library.Relation("construct_datetime_ms_tz", [
    Field.input("year", Integer), Field.input("month", Integer), Field.input("day", Integer),
    Field.input("hour", Integer), Field.input("minute", Integer), Field.input("second", Integer), Field.input("milliseconds", Integer),
    Field.input("timezone", String),
    Field("datetime", DateTime)]
)
_parse_date = library.Relation("parse_date", [Field.input("date_string", String), Field.input("format", String), Field("date", Date)])

# Formatting
_date_format = library.Relation("date_format", [Field.input("date", Date), Field.input("format", String), Field("result", String)])

# Extractors
_date_year = library.Relation("date_year", [Field.input("date", Date), Field("year", Number.size(19, 0))])
_date_quarter = library.Relation("date_quarter", [Field.input("date", Date), Field("quarter", Integer)])
_date_month = library.Relation("date_month", [Field.input("date", Date), Field("month", Integer)])
_date_week = library.Relation("date_week", [Field.input("date", Date), Field("week", Integer)])
_date_day = library.Relation("date_day", [Field.input("date", Date), Field("day", Integer)])
_date_dayofyear = library.Relation("date_dayofyear", [Field.input("date", Date), Field("dayofyear", Integer)])
_date_weekday = library.Relation("date_weekday", [Field.input("date", Date), Field("weekday", Integer)])

# Date Operations
_date_add = library.Relation("date_add", [Field.input("date", Date), Field.input("period", Integer), Field("result", Date)])
_date_subtract = library.Relation("date_subtract", [Field.input("date", Date), Field.input("period", Integer), Field("result", Date)])

# Date Ranges
#   Still unimplemented because we need to add support for emitters to transform lookups using DSL itself. Also,
#   the API is not clear, it could be better to have date_range_from_date where periods can be negative to go backwards.
# _date_range = library.Relation("date_range", [Field.input("start", Date), Field.input("end", Date), Field.input("frequency", String), Field("date", Date)])
# _date_range_from_start = library.Relation("date_range_from_start", [Field.input("start", Date), Field.input("periods", Integer), Field.input("frequency", String), Field("date", Date)])
# _date_range_from_end = library.Relation("date_range_from_end", [Field.input("end", Date), Field.input("periods", Integer), Field.input("frequency", String), Field("date", Date)])

class date:
    def __new__(cls, year: IntegerValue, month: IntegerValue, day: IntegerValue) -> Expression:
        return _construct_date(year, month, day)

    @classmethod
    def year(cls, date: DateValue) -> Expression:
        return _date_year(date)

    @classmethod
    def quarter(cls, date: DateValue) -> Expression:
        return _date_quarter(date)

    @classmethod
    def month(cls, date: DateValue) -> Expression:
        return _date_month(date)

    @classmethod
    def week(cls, date: DateValue) -> Expression:
        return _date_week(date)

    @classmethod
    def day(cls, date: DateValue) -> Expression:
        return _date_day(date)

    @classmethod
    def dayofyear(cls, date: DateValue) -> Expression:
        return _date_dayofyear(date)

    @classmethod
    def isoweekday(cls, date: DateValue) -> Expression:
        """
        Return the ISO weekday as an integer, where Monday is 1, and Sunday is 7.
        """
        return _date_weekday(date)

    @classmethod
    def weekday(cls, date: DateValue) -> Expression:
        return cls.isoweekday(date) - 1 # Convert ISO weekday (1=Mon..7=Sun) to weekday (0=Mon..6=Sun)

    @classmethod
    def fromordinal(cls, ordinal: IntegerValue) -> Expression:
        # ordinal 1 = '0001-01-01'. Minus 1 day since we can't declare date 0000-00-00
        return cls.add(Date(dt.date(1, 1, 1)), days(ordinal - 1))

    @classmethod
    def to_datetime(cls, date: DateValue, hour: int = 0, minute: int = 0, second: int = 0, millisecond: int = 0, tz: str = "UTC") -> Expression:
        _year = cls.year(date)
        _month = cls.month(date)
        _day = cls.day(date)
        return _construct_datetime_ms_tz(_year, _month, _day, hour, minute, second, millisecond, tz)

    @classmethod
    def format(cls, date: DateValue, format: StringValue) -> Expression:
        return _date_format(date, format)

    @classmethod
    def add(cls, date: DateValue, period: Variable) -> Expression:
        return _date_add(date, period)

    @classmethod
    def subtract(cls, date: DateValue, period: Variable) -> Expression:
        return _date_subtract(date, period)

    @classmethod
    def range(cls, start: DateValue | None = None, end: DateValue | None = None, periods: IntegerValue = 1, freq: Frequency = "D") -> Variable:
        if start is None and end is None:
            raise ValueError("Invalid start/end date for date.range. Must provide at least start date or end date")
        if freq not in _days.keys():
            raise ValueError(f"Frequency '{freq}' is not allowed for date_range. List of allowed frequencies: {list(_days.keys())}")
        """
        Note on date_ranges and datetime_range: The way the computation works is that it first overapproximates the
        number of periods.

        For example, date_range(2025-02-01, 2025-03-01, freq='M') and date_range(2025-02-01, 2025-03-31, freq='M') will
        compute range_end to be ceil(28*1/(365/12))=1 and ceil(58*1/(365/12))=2.

        Then, the computation fetches range_end+1 items into _date, which is the right number in the first case but
        one too many in the second case. That's why a filter end >= _date (or variant of) is applied, to remove any
        extra item. The result is two items in both cases.
        """
        # TODO - this transformation is currently LQP-focused. Eventually we will want to
        # move it into the LQP stack and have something general here.
        date_func = cls.add
        if start is None:
            # compute end - periods*freq
            start = end
            end = None
            date_func = cls.subtract
        assert start is not None
        if end is not None:
            num_days = cls.period_days(start, end)
            if freq in ["W", "M", "Y"]:
                range_end = math.ceil(num_days * _days[freq])
                range_end = core.cast(MetaRef(Integer), range_end)
            else:
                range_end = num_days
            # date_range is inclusive. add 1 since std.range is exclusive
            ix = common.range(0, range_end + 1, 1)
        else:
            ix = common.range(0, periods, 1)
        _date = date_func(start, _periods[freq](ix))
        if isinstance(end, dt.date) :
            return select(_date).where(Date(end) >= _date)
        elif end is not None:
            return select(_date).where(end >= _date)
        return _date


    @classmethod
    def period_days(cls, start: DateValue, end: DateValue) -> Expression:
        return _dates_period_days(start, end)

    @classmethod
    def fromisoformat(cls, date_string: StringValue) -> Expression:
        return _parse_date(date_string, ISO.DATE)

#--------------------------------------------------
# DateTime
#--------------------------------------------------

# Constructors
_construct_datetime_ms_tz = library.Relation("construct_datetime_ms_tz", [
    Field.input("year", Integer), Field.input("month", Integer), Field.input("day", Integer),
    Field.input("hour", Integer), Field.input("minute", Integer), Field.input("second", Integer), Field.input("milliseconds", Integer),
    Field.input("timezone", String),
    Field("datetime", DateTime)]
)
_datetime_now = library.Relation("datetime_now", [Field("datetime", DateTime)])
_parse_datetime = library.Relation("parse_datetime", [Field.input("datetime_string", String), Field.input("format", String), Field("datetime", DateTime)])

# Formatting
_datetime_format = library.Relation("datetime_format", [Field.input("datetime", DateTime), Field.input("format", String), Field.input("timezone", String), Field("result", String)])

# Extractors
_datetime_year = library.Relation("datetime_year", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("year", Integer)])
_datetime_quarter = library.Relation("datetime_quarter", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("quarter", Integer)])
_datetime_month = library.Relation("datetime_month", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("month", Integer)])
_datetime_week = library.Relation("datetime_week", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("week", Integer)])
_datetime_day = library.Relation("datetime_day", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("day", Integer)])
_datetime_dayofyear = library.Relation("datetime_dayofyear", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("dayofyear", Integer)])
_datetime_hour = library.Relation("datetime_hour", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("hour", Integer)])
_datetime_minute = library.Relation("datetime_minute", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("minute", Integer)])
# no timezone needed for second extraction
_datetime_second = library.Relation("datetime_second", [Field.input("datetime", DateTime), Field("second", Integer)])
_datetime_weekday = library.Relation("datetime_weekday", [Field.input("datetime", DateTime), Field.input("timezone", String), Field("weekday", Integer)])

# DateTime Operations
_datetime_add = library.Relation("datetime_add", [Field.input("datetime", DateTime), Field.input("period", Integer), Field("result", DateTime)])
_datetime_subtract = library.Relation("datetime_subtract", [Field.input("datetime", DateTime), Field.input("period", Integer), Field("result", DateTime)])

# DateTime Ranges (see comment on Date Ranges)
# _datetime_range = library.Relation("datetime_range", [Field.input("start", DateTime), Field.input("end", DateTime), Field.input("frequency", String), Field("datetime", DateTime)])
# _datetime_range_from_start = library.Relation("datetime_range_from_start", [Field.input("start", DateTime), Field.input("periods", Integer), Field.input("frequency", String), Field("datetime", DateTime)])
# _datetime_range_from_end = library.Relation("datetime_range_from_end", [Field.input("end", DateTime), Field.input("periods", Integer), Field.input("frequency", String), Field("datetime", DateTime)])

class datetime:

    def __new__(cls, year: IntegerValue, month: IntegerValue, day: IntegerValue, hour: IntegerValue = 0, minute: IntegerValue = 0,
             second: IntegerValue = 0, millisecond: IntegerValue = 0, tz: dt.tzinfo|StringValue = "UTC") -> Expression:
        if isinstance(tz, dt.tzinfo):
            tz = str(tz)
        return _construct_datetime_ms_tz(year, month, day, hour, minute, second, millisecond, tz)

    @classmethod
    def now(cls) -> Expression:
        return _datetime_now()

    @classmethod
    def year(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_year(datetime, tz)

    @classmethod
    def quarter(cls, datetime: DateTimeValue,  tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_quarter(datetime, tz)

    @classmethod
    def month(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_month(datetime, tz)

    @classmethod
    def week(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_week(datetime, tz)

    @classmethod
    def day(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_day(datetime, tz)

    @classmethod
    def dayofyear(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_dayofyear(datetime, tz)

    @classmethod
    def hour(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_hour(datetime, tz)

    @classmethod
    def minute(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _datetime_minute(datetime, tz)

    @classmethod
    def second(cls, datetime: DateTimeValue) -> Expression:
        return _datetime_second(datetime)

    @classmethod
    def isoweekday(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        """
        Return the ISO weekday as an integer, where Monday is 1, and Sunday is 7.
        """
        tz = _extract_tz(datetime, tz)
        return _datetime_weekday(datetime, tz)

    @classmethod
    def weekday(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        return cls.isoweekday(datetime, tz) - 1 # Convert ISO weekday (1=Mon..7=Sun) to weekday (0=Mon..6=Sun)

    @classmethod
    def fromordinal(cls, ordinal: IntegerValue) -> Expression:
        # Convert ordinal to milliseconds, since ordinals in Python are days
        # Minus 1 day since we can't declare date 0000-00-00
        ordinal_milliseconds = (ordinal - 1) * 86400000 # 24 * 60 * 60 * 1000
        return cls.add(DateTime(dt.datetime(1, 1, 1, 0, 0, 0)), milliseconds(ordinal_milliseconds))

    @classmethod
    def strptime(cls, date_str: StringValue, format: StringValue) -> Expression:
        return _parse_datetime(date_str, format)

    @classmethod
    def to_date(cls, datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(datetime, tz)
        return _construct_date_from_datetime(datetime, tz)

    @classmethod
    def format(cls, date: DateTimeValue, format: StringValue, tz: dt.tzinfo|StringValue|None = None) -> Expression:
        tz = _extract_tz(date, tz)
        return _datetime_format(date, format, tz)

    @classmethod
    def add(cls, date: DateTimeValue, period: Variable) -> Expression:
        return _datetime_add(date, period)

    @classmethod
    def subtract(cls, date: DateTimeValue, period: Variable) -> Expression:
        return _datetime_subtract(date, period)

    @classmethod
    def range(cls, start: DateTimeValue | None = None, end: DateTimeValue | None = None, periods: IntegerValue = 1, freq: Frequency = "D") -> Variable:
        if start is None and end is None:
            raise ValueError("Invalid start/end date for datetime.range. Must provide at least start date or end date")
        # TODO - this transformation is currently LQP-focused. Eventually we will want to
        # move it into the LQP stack and have something general here.
        _milliseconds = {
            "ms": 1,
            "s": 1 / 1_000,
            "m": 1 / 60_000,
            "H": 1 / 3_600_000,
            "D": 1 / 86_400_000,
            "W": 1 / (86_400_000 * 7),
            "M": 1 / (86_400_000 * (365 / 12)),
            "Y": 1 / (86_400_000 * 365),
        }
        date_func = cls.add
        if start is None:
            start = end
            end = None
            date_func = cls.subtract
        assert start is not None
        if end is not None:
            num_ms = cls.period_milliseconds(start, end)
            if freq == "ms":
                _end = num_ms
            else:
                _end = math.ceil(num_ms * Float(_milliseconds[freq]))
            _end = core.cast(MetaRef(Integer), _end)
            # datetime_range is inclusive. add 1 since common.range is exclusive
            ix = common.range(0, _end + 1, 1)
        else:
            ix = common.range(0, periods, 1)
        _date = date_func(start, _periods[freq](ix))
        if isinstance(end, dt.datetime) :
            return select(_date).where(DateTime(end) >= _date)
        elif end is not None:
            return select(_date).where(end >= _date)
        return _date



    @classmethod
    def period_milliseconds(cls, start: DateTimeValue, end: DateTimeValue) -> Expression:
        return _datetimes_period_milliseconds(start, end)


#--------------------------------------------------
# Periods
#--------------------------------------------------

# Concepts
Nanoseconds = library.Type("Nanoseconds", [Integer])
Microseconds = library.Type("Microseconds", [Integer])
Milliseconds = library.Type("Milliseconds", [Integer])
Seconds = library.Type("Seconds", [Integer])
Minutes = library.Type("Minutes", [Integer])
Hours = library.Type("Hours", [Integer])
Days = library.Type("Days", [Integer])
Weeks = library.Type("Weeks", [Integer])
Months = library.Type("Months", [Integer])
Years = library.Type("Years", [Integer])

# Constructors from Date/DateTime
_dates_period_days = library.Relation("dates_period_days", [Field.input("start", Date), Field.input("end", Date), Field("days", Days)])
_datetimes_period_milliseconds = library.Relation("datetimes_period_milliseconds", [Field.input("start", DateTime), Field.input("end", DateTime), Field("milliseconds", Milliseconds)])

# Basic Constructors
_nanosecond = library.Relation("nanosecond", [Field.input("nanoseconds", Integer), Field("period", Nanoseconds)])
_microsecond = library.Relation("microsecond", [Field.input("microseconds", Integer), Field("period", Microseconds)])
_millisecond = library.Relation("millisecond", [Field.input("milliseconds", Integer), Field("period", Milliseconds)])
_second = library.Relation("second", [Field.input("seconds", Integer), Field("period", Seconds)])
_minute = library.Relation("minute", [Field.input("minutes", Integer), Field("period", Minutes)])
_hour = library.Relation("hour", [Field.input("hours", Integer), Field("period", Hours)])
_day = library.Relation("day", [Field.input("days", Integer), Field("period", Days)])
_week = library.Relation("week", [Field.input("weeks", Integer), Field("period", Weeks)])
_month = library.Relation("month", [Field.input("months", Integer), Field("period", Months)])
_year = library.Relation("year", [Field.input("years", Integer), Field("period", Years)])

def nanoseconds(period: IntegerValue) -> Expression:
    return _nanosecond(period)

def microseconds(period: IntegerValue) -> Expression:
    return _microsecond(period)

def milliseconds(period: IntegerValue) -> Expression:
    return _millisecond(period)

def seconds(period: IntegerValue) -> Expression:
    return _second(period)

def minutes(period: IntegerValue) -> Expression:
    return _minute(period)

def hours(period: IntegerValue) -> Expression:
    return _hour(period)

def days(period: IntegerValue) -> Expression:
    return _day(period)

def weeks(period: IntegerValue) -> Expression:
    return _week(period)

def months(period: IntegerValue) -> Expression:
    return _month(period)

def years(period: IntegerValue) -> Expression:
    return _year(period)


Frequency = Union[
    Literal["ms"],
    Literal["s"],
    Literal["m"],
    Literal["H"],
    Literal["D"],
    Literal["W"],
    Literal["M"],
    Literal["Y"],
]

_periods = {
    "ms": milliseconds,
    "s": seconds,
    "m": minutes,
    "H": hours,
    "D": days,
    "W": weeks,
    "M": months,
    "Y": years,
}

_days = {
    "D": 1,
    "W": Float(1/7),
    "M": Float(1/(365/12)),
    "Y": Float(1/365),
}

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def _extract_tz(datetime: DateTimeValue, tz: dt.tzinfo|StringValue|None) -> StringValue:
    default_tz = "UTC"
    if tz is None:
        if isinstance(datetime, dt.datetime):
            tz = datetime.tzname() or default_tz
        else:
            tz = default_tz
    elif isinstance(tz, dt.tzinfo) :
        tz = tz.tzname(None) or default_tz
    return tz



__version__ = "0.0.0"
import warnings

from v0.relationalai.semantics.reasoners.graph.core import Graph

__all__ = ["Graph"]

warnings.warn(
    "relationalai.early_access.graphs.* is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.graph.*",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations

from v0.relationalai.semantics.internal import internal as b
from .std import _make_expr

# Coerce a number to Int64.
def int64(value: b.Producer|int) -> b.ConceptMember:
    return b.ConceptMember(b.Int64, value, {})

# Coerce a number to Int128.
def int128(value: b.Producer|int) -> b.ConceptMember:
    return b.ConceptMember(b.Int128, value, {})

def parse_int64(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_int64", value, b.Int64.ref("res"))

def parse_int128(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_int128", value, b.Int128.ref("res"))

# Alias parse_int128 to parse
def parse(value: b.Producer|str) -> b.Expression:
    return parse_int128(value)



from .discharge_constraints import DischargeConstraints
from .dnf_union_splitter import DNFUnionSplitter
from .extract_nested_logicals import ExtractNestedLogicals
from .flatten import Flatten
from .format_outputs import FormatOutputs

__all__ = ["DischargeConstraints", "DNFUnionSplitter", "ExtractNestedLogicals", "Flatten", "FormatOutputs"]



import queue
import threading
import time

from logging import Handler, LogRecord
import traceback
from uuid import UUID

from v0.relationalai.util.span_tracker import SpanTracker, record_is_span


class SnowflakeHandler(Handler):
    """
    This is a logging handler that inserts spans directly into a Snowflake table.

    It uses a queue and a worker thread to buffer spans and then insert them in batches.
    """

    def __init__(self, trace_id: UUID, snowflake_conn, sleep_interval_s=2):
        super().__init__()
        self.span_tracker = SpanTracker(trace_id, span_types_to_skip=None, log_span_attrs_as_str=True)
        Handler.__init__(self)
        self.snowflake_conn = snowflake_conn
        self.queue = queue.Queue()
        self.is_shut_down = False
        self.sleep_interval_s = sleep_interval_s
        self.worker_thread = threading.Thread(target=self._consume_loop)
        self.worker_thread.start()
        print(f'snowflake logger started with trace id {trace_id}')

    def emit(self, record: LogRecord):
        if record_is_span(record):
            self.queue.put(record.msg)

    def _consume_loop(self):
        while True:
            while not self.queue.empty():
                try:
                    batch = self._get_batch()
                    self._send_batch(batch)
                except Exception as e:
                    print('snowflake logger error:', e)
                    traceback.print_exc()
            if self.is_shut_down:
                return
            time.sleep(self.sleep_interval_s)

    def _get_batch(self):
        batch = []
        try:
            while True:
                item = self.queue.get_nowait()
                batch.append(item)
        except queue.Empty:
            pass
        return batch

    def _send_batch(self, batch):
        if not batch:
            return

        inserts = []

        for event in batch:
            if event["event"] == "span_start":
                self.span_tracker.handle_span_start(event)
            elif event["event"] == "span_end":
                span_end_result = self.span_tracker.handle_span_end(event)
                if span_end_result:
                    formatted_span_msg = span_end_result[0]
                    if formatted_span_msg:
                        inserts.append(formatted_span_msg)
        # execute
        if inserts:
            self.snowflake_conn.cursor().executemany(
                """
                INSERT INTO spans_raw (id, parent_id, trace_id, type, key, start_ts, finish_ts, attrs)
                VALUES (%(id)s, %(parent_id)s, %(trace_id)s, %(type)s, %(key)s, %(start_timestamp)s, %(end_timestamp)s, %(attrs)s)
                """,
                inserts,
            )

    
    def shut_down(self):
        self.is_shut_down = True
        print('snowflake logger: waiting to shut down...')
        self.worker_thread.join()
        print('snowflake logger: shut down')



from __future__ import annotations

from v0.relationalai.semantics.metamodel import builtins
from v0.relationalai.semantics.metamodel.ir import Node, Model, Require
from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel.rewrite.discharge_constraints import (
    DischargeConstraintsVisitor
)
from v0.relationalai.semantics.lqp.rewrite.functional_dependencies import (
    is_valid_unique_constraint, normalized_fd
)

_DISABLE_CONSTRAINT_DECLARATIONS = True

class AnnotateConstraints(Pass):
    """
    Extends `DischargeConstraints` pass by discharging only those Require nodes that cannot
    be declared as constraints in LQP.

    More precisely, the pass annotates Require nodes depending on how they should be
    treated when generating code:
     * `@declare_constraint` if the Require represents a constraint that can be declared in LQP.
     * `@discharge` if the Require represents a constraint that should be dismissed during
       code generation. Namely, when it cannot be declared in LQP and uses one of the
       `unique`, `exclusive`, `anyof` builtins. These nodes are removed from the IR model
       in the Flatten pass.
    """

    def rewrite(self, model: Model, options: dict = {}) -> Model:
        return AnnotateConstraintsRewriter().walk(model)


class AnnotateConstraintsRewriter(DischargeConstraintsVisitor):
    """
    Visitor marks all nodes which should be removed from IR model with `discharge` annotation.
    """

    def _should_be_declarable_constraint(self, node: Require) -> bool:
        if _DISABLE_CONSTRAINT_DECLARATIONS:
            return False
        if not is_valid_unique_constraint(node):
            return False
        # Currently, we only declare non-structural functional dependencies.
        fd = normalized_fd(node)
        assert fd is not None  # already checked by _is_valid_unique_constraint
        return not fd.is_structural

    def handle_require(self, node: Require, parent: Node):
        if self._should_be_declarable_constraint(node):
            return node.reconstruct(
                node.engine,
                node.domain,
                node.checks,
                node.annotations | [builtins.declare_constraint_annotation]
            )

        return super().handle_require(node, parent)



from v0.relationalai.early_access.dsl.core.exprs import Expr, contextStack
from v0.relationalai.early_access.dsl.core.exprs.scalar import box_number


# Relational non-scalar Exprs

class RelationVariable(Expr):

    # RelationVariables may be "invoked" to declare Atoms
    #
    def __call__(self, *args):
        return contextStack.root_context().build_atom(self, [box_number(a) for a in args])

    def arity(self): pass

    def typeof(self): pass

    def variable(self): return True



from __future__ import annotations

from v0.relationalai.semantics.metamodel import ir
from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel.visitor import Visitor, Rewriter
from v0.relationalai.semantics.metamodel.util import OrderedSet
from v0.relationalai.semantics.metamodel import helpers, factory as f
from typing import Optional, Any

"""
Handle DNF decomposition of unions

E.g., we go from

Logical
    Logical
        Union
            Logical
                Foo(foo::Foo)
                a(foo::Foo, a::Int128)
                a::Int128 < 100::Int128
                Union
                    Logical
                        b(foo::Foo, b::Int128)
                        b::Int128 < 100::Int128
                    Logical
                        c(foo::Foo, c::Int128)
                        c::Int128 > 0::Int128
            Logical
                Foo(foo::Foo)
                id(foo::Foo, id::Int128)
                id::Int128 < 100::Int128
            Logical
                Foo(foo::Foo)
                x(foo::Foo, x::Int128)
                x::Int128 < 100::Int128
                Union
                    Logical
                        y(foo::Foo, y::Int128)
                        y::Int128 < 100::Int128
                    Logical
                        z(foo::Foo, z::Int128)
                        z::Int128 > 0::Int128
        ...
        -> output(...)

to

Logical
    Logical
        Logical
            Foo(foo::Foo)
            a(foo::Foo, a::Int128)
            a::Int128 < 100::Int128
            Logical
                b(foo::Foo, b::Int128)
                b::Int128 < 100::Int128
        ...
        -> output(...)
    Logical
        Logical
            Foo(foo::Foo)
            a(foo::Foo, a::Int128)
            a::Int128 < 100::Int128
            Logical
                c(foo::Foo, c::Int128)
                c::Int128 > 0::Int128
        ...
        -> output(...)
    Logical
        Logical
            Foo(foo::Foo)
            id(foo::Foo, id::Int128)
            id::Int128 < 100::Int128
        ...
        -> output(...)
    Logical
        Logical
            Foo(foo::Foo)
            x(foo::Foo, x::Int128)
            x::Int128 < 100::Int128
            Logical
                y(foo::Foo, y::Int128)
                y::Int128 < 100::Int128
        ...
        -> output(...)
    Logical
        Logical
            Foo(foo::Foo)
            x(foo::Foo, x::Int128)
            x::Int128 < 100::Int128
            Logical
                z(foo::Foo, z::Int128)
                z::Int128 > 0::Int128
        ...
        -> output(...)
"""
class DNFUnionSplitter(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        visitor = DNFExtractor()
        model.accept(visitor)
        return DNFRewriter(visitor).walk(model) if visitor.replaced_by else model

class DNFExtractor(Visitor):
    def __init__(self):
        # The logical that contains the output.
        # The assumption for the IR at this point is that there is only one output.
        self.output_logical: Optional[ir.Logical] = None
        self.active_negations: list[ir.Not] = []
        # Nodes that have to split into multiple similar nodes, depending on the changes
        # of sub-nodes.
        self.should_split: OrderedSet[ir.Node] = OrderedSet()
        # Track that a node has to be replaced by other nodes.
        # If node X should be replaced by nodes Y and Z, then the parent should be replicated
        # one time replacing X with Y and one time replacing X with Z.
        self.replaced_by: dict[ir.Node, list[ir.Task]] = {}

    def enter(self, node: ir.Node, parent: Optional[ir.Node]=None) -> Visitor:
        if isinstance(node, ir.Logical):
            if any(isinstance(x, ir.Output) for x in node.body):
                assert not self.output_logical, "multiple outputs"
                self.output_logical = node

        elif isinstance(node, ir.Not):
            self.active_negations.append(node)

        return self

    def leave(self, node: ir.Node, parent: Optional[ir.Node]=None) -> ir.Node:
        if isinstance(node, ir.Logical) and node in self.should_split:
            # The given logical, will be replaced by multiple logicals, each with a different
            # group of tasks. We need to generate all the possible groups.
            # A list of logical bodies (lists).
            replacement_bodies: list[list[ir.Task]] = [[]]
            for task in node.body:
                if task in self.replaced_by:
                    new_replacement_bodies: list[list[ir.Task]] = []
                    replacement_tasks = self.replaced_by[task]
                    for body in replacement_bodies:
                        for new_task in replacement_tasks:
                            # copy to mutate
                            new_body = list(body)
                            new_body.append(new_task.clone())
                            new_replacement_bodies.append(new_body)
                    replacement_bodies = new_replacement_bodies

                else:
                    for new_body in replacement_bodies:
                        new_body.append(task.clone())

            replacement_tasks: list[ir.Task] = []
            for body in replacement_bodies:
                new_task = f.logical(body, node.hoisted)
                replacement_tasks.append(new_task)
            self.replaced_by[node] = replacement_tasks

            if node != self.output_logical:
                self.should_split.add(parent)
            elif node == self.output_logical:
                assert parent and isinstance(parent, ir.Logical)
                new_parent = f.logical(tuple(replacement_tasks), node.hoisted)
                self.replaced_by[parent] = [new_parent]

        elif isinstance(node, ir.Not) and self.active_negations[-1] == node:
            self.active_negations.pop()

        elif (isinstance(node, ir.Union) and
              self.output_logical and
              len(self.active_negations) % 2 == 0 and
              len(node.tasks) > 1):
            # We split the union when there is a branch with vars "X,Y" and another with "X,Z"
            # If some branches have vars "X, Y, Z" and others have "X, Y" or "Y, Z" we don't split
            should_split = False
            all_vars = helpers.collect_vars(node.tasks[0])
            for t in node.tasks[1:]:
                vars = helpers.collect_vars(t)
                curr_intersection = vars.get_set().intersection(all_vars.get_set())
                should_split |= not (curr_intersection == vars.get_set() or curr_intersection == all_vars.get_set())
                if should_split:
                    replacements:list[ir.Task] = []
                    for t in node.tasks:
                        # If some branch should already be replaced, we flatten all the replacements here.
                        if t in self.replaced_by:
                            replacements.extend(self.replaced_by[t])
                        else:
                            replacements.append(t)
                    self.replaced_by[node] = replacements
                    self.should_split.add(parent)
                    break
                all_vars.update(vars)

        if isinstance(node, ir.Logical) and node == self.output_logical:
            self.output_logical = None

        return node

class DNFRewriter(Rewriter):
    def __init__(self, visitor: DNFExtractor):
        super().__init__()
        self.visitor = visitor
        self.outer_parent_logical: Optional[ir.Logical] = None

    def handle_logical(self, node: ir.Logical, parent: ir.Node, ctx:Optional[Any]=None) -> ir.Logical:
        if node in self.visitor.replaced_by:
            new_tasks = self.visitor.replaced_by[node]
            assert len(new_tasks) == 1
            new_task = new_tasks[0]
            assert isinstance(new_task, ir.Logical)
            return new_task
        return node



#--------------------------------------------------
# Compiler
#--------------------------------------------------

from v0.relationalai.semantics.lqp import ir as lqp, model2lqp
from v0.relationalai.semantics.lqp.utils import TranslationCtx, UniqueNames
from v0.relationalai.semantics.lqp.passes import lqp_passes
from v0.relationalai.semantics.metamodel import ir, compiler as c

from typing import Optional

class Compiler(c.Compiler):
    def __init__(self):
        super().__init__(lqp_passes())
        self.def_names = UniqueNames()

    def do_compile(self, model: ir.Model, options:dict={}) -> tuple[Optional[tuple], lqp.Epoch]:
        fragment_id: bytes = options.get("fragment_id", bytes(404))
        # Reset the var context for each compilation
        # TODO: Change to unique var names per lookup
        # TODO: Have our needs realigned with NameCache?
        return model2lqp.to_lqp(model, fragment_id, TranslationCtx(self.def_names))



from __future__ import annotations
from typing import Union

from .. import dsl, std, errors, metamodel as m


# Custom types
_String = Union[str, dsl.Producer]
_Integer = Union[int, dsl.Producer]

def escape(regex: _String) -> dsl.Expression:
    return dsl.rel.escape_regex_metachars(regex)

dsl.tag(dsl.rel.pyrel_regex_match, m.Builtins.Filter)
dsl.tag(dsl.rel.pyrel_regex_search, m.Builtins.Filter)
dsl.tag(dsl.rel.regex_compile, m.Builtins.Expensive)
dsl.tag(dsl.rel.num_chars, m.Builtins.SingleValued)

class Pattern:
    def __init__(self, graph: dsl.Graph, regex: _String):
        self._graph = graph
        self.pattern = regex
        with graph.rule():
            dsl.rel.__compiled_patterns.add(regex, dsl.rel.regex_compile(regex))

    @property
    def _compiled(self):
        return dsl.rel.__compiled_patterns(self.pattern)

    def match(self, string: _String, pos: _Integer = 0) -> Match:
        return Match(self._graph, self._compiled, string, pos)

    def search(self, string: _String, pos: _Integer = 0) -> Match:
        return Match(self._graph, self._compiled, string, pos, _how="search")

    def fullmatch(self, string: _String, pos: _Integer = 0) -> Match:
        return Match(self._graph, self._compiled, string, pos, _how="fullmatch")

    def findall(self, string: _String) -> tuple[dsl.Producer, dsl.Producer]:
        return findall(self.pattern, string)

    def sub(self, repl: _String, string: _String) -> dsl.Expression:
        # NOTE: self.pattern must be used here instead of self._compiled
        # because the compiled pattern throws an FFIException.
        return sub(self.pattern, repl, string)

    def _to_var(self):
        raise errors.NonVarObject(self, "Pattern objects cannot be used as variables")


def compile(regex: _String) -> Pattern:
    return Pattern(dsl.get_graph(), regex)


class Match(dsl.Producer):
    def __init__(self, graph:dsl.Graph, regex: Pattern | _String, string: _String, pos: _Integer = 0, _how: str = "match"):
        super().__init__(graph)
        if isinstance(regex, Pattern):
            self.re = regex._compiled
        else:
            self.re = dsl.rel.regex_compile(regex)
        self.string = string
        self.pos = pos
        self._how = _how
        self._fullmatch = dsl.create_var()
        self._start = dsl.create_var()
        if self._how == "match":
            dsl.rel.pyrel_regex_match(self.re, self.string, self.pos + 1, self._start, self._fullmatch)
        elif self._how == "search":
            dsl.rel.pyrel_regex_search(self.re, self.string, self.pos + 1, self._start, self._fullmatch)
        elif self._how == "fullmatch":
            dsl.rel.pyrel_regex_match(self.re, self.string, self.pos + 1, self._start, self._fullmatch)
            self._fullmatch == dsl.rel.substring(self.string, self.pos + 1, dsl.rel.num_chars(self.string))
        else:
            raise ValueError(f"Invalid match type: {self._how!r}")

    def group(self, index: _Integer | _String = 0) -> dsl.Instance:
        if index == 0:
            return self._fullmatch
        else:
            return dsl.rel.pyrel_capture_group(self.re, self.string, self.pos + 1, index)

    def start(self) -> dsl.Expression:
        return self._start - 1

    def end(self) -> dsl.Expression:
        return dsl.rel.num_chars(self._fullmatch) + self._start - 1

    def span(self) -> tuple[dsl.Producer, dsl.Producer]:
        return self.start(), self.end()

    def __getitem__(self, index: _Integer | _String) -> dsl.Instance:
        return self.group(index)

    def _to_var(self):
        return self.group()


def match(regex: _String, string: _String) -> Match:
    return Match(dsl.get_graph(), regex, string)


def search(regex: _String, string: _String) -> Match:
    return Match(dsl.get_graph(), regex, string, _how="search")


def fullmatch(regex: _String, string: _String) -> Match:
    return Match(dsl.get_graph(), regex, string, _how="fullmatch")


def findall(regex: _String, string: _String) -> tuple[dsl.Producer, dsl.Producer]:
    ix, match = dsl.create_vars(2)
    dsl.rel.regex_match_all(regex, string, ix, match)
    rank = std.aggregates.rank_asc(ix, match, per=[string])
    return rank, match


def sub(regex: _String, repl: _String, string: _String) -> dsl.Expression:
    return dsl.rel.pyrel_regex_sub(regex, repl, string)


__all__ = [
    "compile",
    "escape",
    "findall",
    "fullmatch",
    "match",
    "search",
    "sub",
]



import warnings

from v0.relationalai.semantics.metamodel.types import (digits_to_bits, Hash, String, Number, Int64, Int128, Bool, Date,
                                                    DateTime, Float, RowId, UInt128, Any, Symbol, is_subtype,
                                                    is_decimal_subtype, decimal_supertype)

__all__ = ["digits_to_bits", "Int64", "Int128", "DateTime", "Hash", "String", "Number", "Bool", "Date", "Float",
           "RowId", "UInt128", "Any", "Symbol", "is_subtype", "is_decimal_subtype", "decimal_supertype"]

warnings.warn(
    "relationalai.early_access.metamodel.types is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.types",
    DeprecationWarning,
    stacklevel=2,
)


from typing import Optional, Iterable, List, Sequence as _Sequence
from dataclasses import dataclass, field

from .metamodel import *
import datetime as _dt
from ...util.dataclasses import print_tree as print_dataclass_tree
from ...util.naming import NameCache

#------------------------------------------------------
# pprint
#------------------------------------------------------

_INDENT = "    "
_DERIVE_ARROW = "→"
VERBOSE = False
# VERBOSE = True

def _indent_line(indent: int, text: str) -> str:
        return f"{_INDENT * indent}{text}"

@dataclass(frozen=True)
class Printer():
    """Pretty Printer for Metamodel Nodes"""
    name_cache: NameCache = field(default_factory=NameCache)

    def format_type(self, type_: Type, include_supertypes: bool = False) -> str:
        if type_ is None:
            return ""
        if isinstance(type_, ScalarType):
            if include_supertypes and type_.super_types:
                supertypes = ", ".join(
                    self.format_type(t, include_supertypes=True) for t in type_.super_types
                )
                return f"{type_.name or 'Scalar'} < {supertypes}" if supertypes else (type_.name or "Scalar")
            else:
                return type_.name or "Scalar"
        if isinstance(type_, UnionType):
            inner = [name for name in (self.format_type(t) for t in type_.types) if name]
            joined = " | ".join(inner)
            return joined or "Union"
        if isinstance(type_, ListType):
            inner = self.format_type(type_.element_type)
            return f"List[{inner}]" if inner else "List"
        if isinstance(type_, TypeNode):
            return type_.__class__.__name__
        return str(type_)

    def primitive(self, value: Primitive) -> str:
        if value is None:
            return "None"
        if isinstance(value, bool):
            return "true" if value else "false"
        if isinstance(value, str):
            escaped = value.replace("\"", "\\\"")
            return f'"{escaped}"'
        if isinstance(value, _dt.datetime):
            return value.isoformat()
        return str(value)

    def var(self, var: Var) -> str:
        name = var.name or f"var_{var.id}"
        name = self.name_cache.get_name(var.id, name)
        type_name = self.format_type(var.type)
        return f"{name}::{type_name}" if type_name else name

    def literal(self, literal: Literal) -> str:
        value = self.primitive(literal.value)
        type_name = self.format_type(literal.type)
        return f"{value}::{type_name}" if type_name else value

    def binding(self, binding: Var) -> str:
        return self.var(binding)

    def relation(self, relation: Relation) -> str:
        name = relation.name or f"Relation#{relation.id}"
        if VERBOSE:
            fields = ", ".join(
                f"{field.name or field.id}::{self.format_type(field.type)}"
                for field in relation.fields
            )
            return f"{name}[{fields}]"
        else:
            return name

    def capability_str(self, capability: Capability) -> str:
        return capability.name or f"Capability#{capability.id}"

    def reasoner_str(self, reasoner: Reasoner) -> str:
        name = reasoner.type or f"Reasoner#{reasoner.id}"
        suffix_parts: List[str] = []
        if reasoner.capabilities:
            suffix_parts.append(
                "capabilities[" + ", ".join(self.capability_str(cap) for cap in reasoner.capabilities) + "]"
            )
        if reasoner.relations:
            suffix_parts.append(
                "relations[" + ", ".join(self.relation(rel) for rel in reasoner.relations) + "]"
            )
        if reasoner.info is not None:
            suffix_parts.append(f"info={reasoner.info!r}")
        if suffix_parts:
            return f"{name} " + " ".join(suffix_parts)
        return name

    def value(self, val: Value) -> str:
        if isinstance(val, Var):
            return self.var(val)
        if isinstance(val, Literal):
            return self.literal(val)
        if isinstance(val, Relation):
            return self.relation(val)
        if isinstance(val, TypeNode):
            return self.format_type(val)
        if isinstance(val, Field):
            return f"{val._relation.name}.{val.name or val.id}"
        if isinstance(val, list):
            return "[" + ", ".join(self.value(v) for v in val) + "]"
        if isinstance(val, tuple):
            return "(" + ", ".join(self.value(v) for v in val) + ")"
        if isinstance(val, bool):
            return "true" if val else "false"
        if val is None:
            return "None"
        return str(val)


    def annotation(self, annotation: Annotation) -> str:
        name = self.relation(annotation.relation)
        if not annotation.args:
            return f"@{name}"
        args = ", ".join(self.value(arg) for arg in annotation.args)
        return f"@{name}({args})"


    def header(
        self,
        name: str,
        *,
        annotations: Optional[_Sequence[Annotation]] = None,
        suffix: str = "",
    ) -> str:
        parts = [name]
        if suffix:
            parts.append(suffix)
        head = " ".join(part for part in parts if part).rstrip()
        if annotations:
            annos = " ".join(self.annotation(anno) for anno in annotations)
            head = f"{head} {annos}" if annos else head
        return head


    def relation_call(self, rel: Relation, args: _Sequence[Value], reading: Reading|None) -> str:
        args_str = ", ".join(self.value(arg) for arg in args)
        reading_ix = rel.readings.index(reading) if reading else 0
        reading_suffix = f"<{reading_ix}>" if reading_ix > 0 else ""
        return f"{self.relation(rel)}({args_str}){reading_suffix}"


    def _render_task(self, task: Task, indent: int = 0) -> Iterable[str]:
        if isinstance(task, Container):
            head = f"{type(task).__name__}{'{…}' if task.scope else ''}{'?' if task.optional else ''}"
        else:
            head = ""

        if isinstance(task, Logical):
            yield _indent_line(
                indent,
                self.header(
                    head, annotations=task.annotations
                ),
            )
            for inner in task.body:
                yield from self._render_task(inner, indent + 1)
            return

        if isinstance(task, Sequence):
            yield _indent_line(
                indent,
                self.header(
                    head, annotations=task.annotations
                ),
            )
            for inner in task.tasks:
                yield from self._render_task(inner, indent + 1)
            return

        if isinstance(task, Union):
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations),
            )
            for inner in task.tasks:
                yield from self._render_task(inner, indent + 1)
            return

        if isinstance(task, Match):
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations),
            )
            for inner in task.tasks:
                yield from self._render_task(inner, indent + 1)
            return

        if isinstance(task, Until):
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations),
            )
            yield _indent_line(indent + 1, "check")
            yield from self._render_task(task.check, indent + 2)
            yield _indent_line(indent + 1, "body")
            yield from self._render_task(task.body, indent + 2)
            return

        if isinstance(task, Wait):
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations),
            )
            yield _indent_line(indent + 1, "check")
            yield from self._render_task(task.check, indent + 2)
            return

        if isinstance(task, Loop):
            suffix_parts: List[str] = []
            if task.over:
                suffix_parts.append(
                    "over [" + ", ".join(self.var(v) for v in task.over) + "]"
                )
            if task.concurrency != 1:
                suffix_parts.append(f"concurrency={task.concurrency}")
            suffix = " ".join(suffix_parts)
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations, suffix=suffix),
            )
            yield _indent_line(indent + 1, "body")
            yield from self._render_task(task.body, indent + 2)
            return

        if isinstance(task, Break):
            yield _indent_line(
                indent, self.header("Break", annotations=task.annotations)
            )
            yield _indent_line(indent + 1, "check")
            yield from self._render_task(task.check, indent + 2)
            return

        if isinstance(task, Require):
            yield _indent_line(
                indent,
                self.header(
                    head, annotations=task.annotations
                ),
            )
            yield _indent_line(indent + 1, "domain")
            if not isinstance(task.domain, Logical) or task.domain.body:
                yield from self._render_task(task.domain, indent + 2)
            yield _indent_line(indent + 1, "check")
            yield from self._render_task(task.check, indent + 2)
            if task.error:
                yield _indent_line(indent + 1, "error")
                yield from self._render_task(task.error, indent + 2)
            return

        if isinstance(task, Not):
            yield _indent_line(
                indent, self.header(head, annotations=task.annotations)
            )
            yield from self._render_task(task.task, indent + 1)
            return

        if isinstance(task, Exists):
            suffix = ""
            if task.vars:
                suffix = "[" + ", ".join(self.var(v) for v in task.vars) + "]"
            yield _indent_line(
                indent,
                self.header(head, annotations=task.annotations, suffix=suffix),
            )
            yield from self._render_task(task.task, indent + 1)
            return

        if isinstance(task, Lookup):
            yield _indent_line(
                indent,
                self.header(
                    self.relation_call(task.relation, task.args, task.reading_hint),
                    annotations=task.annotations,
                ),
            )
            return

        if isinstance(task, Update):
            yield _indent_line(
                indent,
                self.header(
                    f"{_DERIVE_ARROW} {task.effect.value} "
                    + self.relation_call(task.relation, task.args, task.reading_hint),
                    annotations=task.annotations,
                ),
            )
            return

        if isinstance(task, Aggregate):
            args = ", ".join(self.value(arg) for arg in task.args)
            suffix_parts: List[str] = []
            if task.projection:
                suffix_parts.append(
                    "projection["
                    + ", ".join(self.var(v) for v in task.projection)
                    + "]"
                )
            if task.group:
                suffix_parts.append(
                    "group[" + ", ".join(self.var(v) for v in task.group) + "]"
                )
            suffix = " ".join(suffix_parts)
            head = f"aggregate {self.relation(task.aggregation)}({args})"
            if suffix:
                head = f"{head} {suffix}"
            yield _indent_line(indent, self.header(head, annotations=task.annotations))
            if task.body:
                yield from self._render_task(task.body, indent + 1)
            return

        if isinstance(task, Construct):
            values = [self.value(val) for val in task.values]
            id_var = task.id_var
            if isinstance(id_var, Var):
                has_identity = id_var.name or getattr(id_var, "type", None) is not None
                if has_identity:
                    values.append(self.var(id_var))
            elif id_var is not None:
                values.append(self.value(id_var))
            line = f"construct({', '.join(values)})"
            yield _indent_line(
                indent, self.header(line, annotations=task.annotations)
            )
            return

        yield _indent_line(indent, object.__repr__(task))

    def _render_model(self, model: Model, indent: int = 0) -> Iterable[str]:
        summary_parts: List[str] = []
        if model.reasoners:
            summary_parts.append(f"reasoners={len(model.reasoners)}")
        if model.relations:
            summary_parts.append(f"relations={len(model.relations)}")
        if model.types:
            summary_parts.append(f"types={len(model.types)}")
        head = "Model"
        if summary_parts:
            head = f"{head} ({', '.join(summary_parts)})"
        yield _indent_line(indent, head)
        if model.reasoners:
            yield _indent_line(indent + 1, "reasoners")
            for reasoner in model.reasoners:
                yield _indent_line(indent + 2, self.reasoner_str(reasoner))
        if model.relations:
            yield _indent_line(indent + 1, "relations")
            for rel in model.relations:
                yield _indent_line(indent + 2, self.relation(rel))
        if model.types:
            yield _indent_line(indent + 1, "types")
            for type_ in model.types:
                yield _indent_line(indent + 2, self.format_type(type_, include_supertypes=VERBOSE))
        yield _indent_line(indent + 1, "root")
        if isinstance(model.root, Task):
            yield from self._render_task(model.root, indent + 2)
        else:
            yield _indent_line(indent + 2, object.__repr__(model.root))

    def format_task(self, task: Task, indent: Optional[int] = None) -> str:
        start_indent = 0 if indent is None else indent
        return "\n".join(self._render_task(task, start_indent))

    def format_model(self, model: Model, indent: Optional[int] = None) -> str:
        start_indent = 0 if indent is None else indent
        return "\n".join(self._render_model(model, start_indent))

    def format(self, node: Node|list|tuple, indent: Optional[int] = None) -> str:
        start_indent = 0 if indent is None else indent
        if isinstance(node, Task):
            return self.format_task(node, start_indent)
        elif isinstance(node, Model):
            return self.format_model(node, start_indent)
        elif isinstance(node, TypeNode):
            return self.format_type(node)
        elif isinstance(node, Literal):
            return self.literal(node)
        elif isinstance(node, Var):
            return self.var(node)
        elif isinstance(node, Relation):
            return self.relation(node)
        elif isinstance(node, (list, tuple)):
            return "List\n" + "\n".join(self.format(n, start_indent+1) for n in node)
        elif isinstance(node, Field):
            return self.value(node)
        elif isinstance(node, Annotation):
            return self.annotation(node)
        else:
            return object.__repr__(node)


def format(node: Node|list|tuple, indent: Optional[int] = None) -> str:
    return Printer().format(node, indent)

def pprint(node: Node|list|tuple, indent: Optional[int] = None):
    print(Printer().format(node, indent))

def print_tree(node: Node):
    print_dataclass_tree(node, hide_fields=["source"])



from .annotate_constraints import AnnotateConstraints
from .cdc import CDC
from .extract_common import ExtractCommon
from .extract_keys import ExtractKeys
from .function_annotations import FunctionAnnotations, SplitMultiCheckRequires
from .quantify_vars import QuantifyVars
from .splinter import Splinter

__all__ = [
    "AnnotateConstraints",
    "CDC",
    "ExtractCommon",
    "ExtractKeys",
    "FunctionAnnotations",
    "QuantifyVars",
    "Splinter",
    "SplitMultiCheckRequires",
]



"""
RelationalAI CLI - Tools for working with semantic models.
For more documentation on these commands, visit: https://docs.relationalai.com
"""

from __future__ import annotations

from pathlib import Path

import click

from rich.console import Console

from .config_template import CONFIG_TEMPLATE

console = Console()


@click.group(invoke_without_command=True)
@click.version_option()
@click.pass_context
def cli(ctx: click.Context):
    """RelationalAI CLI - Tools for working with semantic models. For more documentation on these commands, visit: https://docs.relationalai.com"""
    if ctx.invoked_subcommand is None:
        # Show help if no command provided
        click.echo(ctx.get_help())


@cli.command()
def analyze():
    """Check for errors, warnings, and performance issues in the models."""
    console.print("analyze")


@cli.command()
def build():
    """Create artifacts, show compilation errors, and produce SQL for the models."""
    console.print("build")


@cli.command()
def connect():
    """Validate config and database connection."""
    console.print("connect")


@cli.command()
def deploy():
    """Deploy models/views to the database and produce SQL for execution."""
    console.print("deploy")


@cli.command()
def explore():
    """Model explorer for visualization of data."""
    console.print("explore")


@cli.command()
def init():
    """Create a template for the YAML config file."""
    config_file = Path("raiconfig.yml")

    if config_file.exists():
        console.print(f"[yellow]Config file '{config_file}' already exists. Skipping creation.[/yellow]")
        return

    try:
        config_file.write_text(CONFIG_TEMPLATE, encoding="utf-8")
        console.print(f"[green]✓ Created config file '{config_file}'[/green]")
    except Exception as e:
        console.print(f"[red]✗ Failed to create config file: {e}[/red]")
        raise click.ClickException(f"Failed to create config file: {e}")


@cli.command()
@click.option("--uninstall", is_flag=True, help="Nukes the database")
def clean(uninstall: bool):
    """Clean up build folder and remove artifacts."""
    console.print("clean")


@cli.command()
def test():
    """Run constraints and tests."""
    console.print("test")


if __name__ == "__main__":
    cli()



"""
CLI Resources - Resources class for CLI operations.
Re-raises exceptions without transformation and provides CLI utility methods.
"""
from __future__ import annotations
import json
from typing import Any

from ....tools.constants import RAI_APP_NAME

# Import Resources from snowflake - this creates a dependency but no circular import
# since snowflake.py doesn't import from this file
from .snowflake import Resources, ExecContext


class CLIResources(Resources):
    """
    Resources class for CLI operations.
    Re-raises exceptions without transformation and provides CLI utility methods.
    """

    def _handle_standard_exec_errors(self, e: Exception, ctx: ExecContext) -> Any | None:
        """For CLI resources, re-raise exceptions without transformation."""
        raise e

    def list_warehouses(self):
        """List all warehouses in the Snowflake account."""
        results = self._exec("SHOW WAREHOUSES")
        if not results:
            return []
        return [{"name": name}
                for (name, *rest) in results]

    def list_compute_pools(self):
        """List all compute pools in the Snowflake account."""
        results = self._exec("SHOW COMPUTE POOLS")
        if not results:
            return []
        return [{"name": name, "status": status, "min_nodes": min_nodes, "max_nodes": max_nodes, "instance_family": instance_family}
                for (name, status, min_nodes, max_nodes, instance_family, *rest) in results]

    def list_roles(self):
        """List all available roles in the Snowflake account."""
        results = self._exec("SELECT CURRENT_AVAILABLE_ROLES()")
        if not results:
            return []
        # the response is a single row with a single column containing
        # a stringified JSON array of role names:
        row = results[0]
        if not row:
            return []
        return [{"name": name} for name in json.loads(row[0])]

    def list_apps(self):
        """List all applications in the Snowflake account."""
        all_apps = self._exec(f"SHOW APPLICATIONS LIKE '{RAI_APP_NAME}'")
        if not all_apps:
            all_apps = self._exec("SHOW APPLICATIONS")
            if not all_apps:
                return []
        return [{"name": name}
                for (time, name, *rest) in all_apps]

    def list_databases(self):
        """List all databases in the Snowflake account."""
        results = self._exec("SHOW DATABASES")
        if not results:
            return []
        return [{"name": name}
                for (time, name, *rest) in results]

    def list_sf_schemas(self, database: str):
        """List all schemas in a given database."""
        results = self._exec(f"SHOW SCHEMAS IN {database}")
        if not results:
            return []
        return [{"name": name}
                for (time, name, *rest) in results]

    def list_tables(self, database: str, schema: str):
        """List all tables and views in a given schema."""
        results = self._exec(f"SHOW OBJECTS IN {database}.{schema}")
        items = []
        if results:
            for (time, name, db_name, schema_name, kind, *rest) in results:
                items.append({"name": name, "kind": kind.lower()})
        return items



from __future__ import annotations
from typing import cast
from decimal import Decimal as PyDecimal
from v0.relationalai.semantics.metamodel import ir, types
from v0.relationalai.semantics.internal import internal as b
from .std import _make_expr

#--------------------------------------------------
# Constructors
#--------------------------------------------------

def decimal(value: b.Producer|int|float|PyDecimal, precision=38, scale=14) -> b.Expression:
    """
    Create an expression that represents a decimal with this value, precision and scale.
    """
    if isinstance(value, int):
        value = PyDecimal(str(value))
    if isinstance(value, float):
        value = PyDecimal(str(value))
    return b.ConceptMember(b.decimal_concept(precision, scale), value, {})

def parse_decimal(value: b.Producer|str, precision=38, scale=14) -> b.Expression:
    """
    Create an expression that represents parsing this string value as a decimal with this
    precision and scale.
    """
    return _make_expr("parse_decimal", value, b.decimal_concept(precision, scale).ref("res"))

def parse(value: b.Producer|str, decimal: b.Concept) -> b.Expression:
    """
    Create an expression that represents parsing this string value as a decimal with the
    precision and scale of the decimal argument.
    """
    return parse_decimal(value, precision(decimal), scale(decimal))

#--------------------------------------------------
# Decimal information.
#--------------------------------------------------

def is_decimal(decimal: b.Concept) -> bool:
    return b.is_decimal(decimal)

def precision(decimal: b.Concept) -> int:
    """ Assuming the concept represents a decimal, return its precision. """
    assert b.is_decimal(decimal)
    typ = cast(ir.DecimalType, types.decimal_by_type_str(decimal._name))
    return typ.precision

def scale(decimal: b.Concept) -> int:
    """ Assuming the concept represents a decimal, return its scale. """
    assert b.is_decimal(decimal)
    typ = cast(ir.DecimalType, types.decimal_by_type_str(decimal._name))
    return typ.scale

def size(decimal: b.Concept) -> int:
    """
    Assuming the concept represents a decimal, return its size, i.e. the number of bits
    needed to represent the decimal.
    """
    assert b.is_decimal(decimal)
    typ = cast(ir.DecimalType, types.decimal_by_type_str(decimal._name))
    return types.digits_to_bits(typ.precision)



from v0.relationalai.semantics.internal import Fragment
from v0.relationalai.semantics.rel import Compiler as RelCompiler
from v0.relationalai.semantics.metamodel import ir
from v0.relationalai.semantics.sql import Compiler as SqlCompiler

from dataclasses import dataclass
from typing import Optional
import inspect
import tempfile
import webbrowser
import os
import html

@dataclass
class CompilationResult:
    python_source: str
    ir: str
    rel: Optional[str]
    sql: Optional[str]

    def show(self):
        """
        Displays the compilation results in a browser by creating a temporary HTML file.

        Args:
            compilation_result: Dictionary containing 'ir', 'rel', and 'sql' keys
                               as returned by the compile() function

        Returns:
            Path to the temporary HTML file
        """

        # Create HTML content with syntax highlighting using highlight.js
        html_content = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>RAI Compilation Results</title>
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/default.min.css">
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/python.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/ruby.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/sql.min.js"></script>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 20px; }}
                .section {{ margin-bottom: 30px; }}
                h2 {{ color: #333; }}
                pre {{ background-color: #f5f5f5; padding: 10px; border-radius: 5px; overflow: auto; }}
                .error {{ color: red; }}
            </style>
        </head>
        <body>
            <h1>RAI Compilation Results</h1>

            <div class="section">
                <h2>Python Source</h2>
                <pre><code class="language-python">{}</code></pre>
            </div>

            <div class="section">
                <h2>IR (Intermediate Representation)</h2>
                <pre><code class="language-json">{}</code></pre>
            </div>

            <div class="section">
                <h2>Rel</h2>
                <pre><code class="language-ruby">{}</code></pre>
            </div>

            <div class="section">
                <h2>SQL</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <script>hljs.highlightAll();</script>
        </body>
        </html>
        """.format(
            html.escape(self.python_source),
            html.escape(self.ir),
            html.escape(self.rel or "not available"),
            html.escape(self.sql or "not available"),
        )

        # Create a temporary HTML file
        with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as f:
            f.write(html_content.encode('utf-8'))
            temp_path = f.name

        # Open the HTML file in the default browser
        webbrowser.open('file://' + os.path.abspath(temp_path))

        return temp_path

class CompilationManager:
    def __init__(self, fragment: Fragment):
        self.fragment = fragment

    def python_source(self):
        """Extract the source code that created this fragment from the caller's file."""
        try:
            # Start with the current frame and work backwards
            frame = inspect.currentframe()

            # Skip the first frame (this method)
            if frame:
                frame = frame.f_back

            relevant_frames = []
            max_frames_to_check = 10  # Limit how far back we go

            # Collect frames that might be relevant
            for i in range(max_frames_to_check):
                if frame is None:
                    break

                filename = frame.f_code.co_filename

                # Skip frames from internal Python modules
                if (not filename.startswith('<') and
                    'site-packages' not in filename and
                    '/lib/python' not in filename):

                    # Check whether this file imports from relationalai.semantics.internal
                    try:
                        with open(filename, 'r') as f:
                            content = f.read()
                            if ('from relationalai.semantics.internal' in content or
                                'import relationalai.semantics.internal' in content):
                                relevant_frames.append(frame)
                    except FileNotFoundError:
                        pass  # If we can't read the file, just continue

                frame = frame.f_back

            # If we found relevant frames, use the last one (furthest up the call stack)
            if relevant_frames:
                frame = relevant_frames[-1]
            else:
                # Fall back to the second frame if we didn't find any relevant ones
                frame = inspect.currentframe()
                for _ in range(2):
                    if frame is None:
                        return "Source code not available"
                    frame = frame.f_back

            if frame is None:
                return "Source code not available"

            # Get the filename and line number
            caller_info = inspect.getframeinfo(frame)
            filename = caller_info.filename
            lineno = caller_info.lineno

            # Read the source file
            if os.path.exists(filename):
                with open(filename, 'r') as f:
                    source_lines = f.readlines()

                # Extract a reasonable context around the line
                start_line = max(0, lineno - 20)
                end_line = min(len(source_lines), lineno + 20)

                # Try to find a complete test method or function
                # Look for the start of a function or method definition before the current line
                for i in range(lineno - 1, max(0, start_line - 10), -1):
                    if i >= len(source_lines):
                        continue
                    line = source_lines[i].strip()
                    if line.startswith('def ') or line.startswith('async def '):
                        start_line = i
                        break

                context_lines = source_lines[start_line:end_line]
                return ''.join(context_lines)

            return "Source file not found"
        except Exception as e:
            return f"Error extracting source code: {str(e)}"

    def ir(self):
        builder_model = self.fragment._model
        if not builder_model:
            raise ValueError("Fragment must be part of a model to compile to IR")
        ir_model = builder_model._to_ir()
        # TODO: do something with the fragment here
        # fragment_task = builder_model._compiler.compile_task(self.fragment)
        return ir_model

    def rel(self, ir_model=None):
        if ir_model is None:
            ir_model = self.ir()
        rel_compiler = RelCompiler()
        full_code = rel_compiler.compile(ir_model)
        return full_code

    def sql(self, ir_model=None):
        if ir_model is None:
            ir_model = self.ir()
        return SqlCompiler().compile(ir_model)

    def compile(self):
        """
        Sends the various compiled forms of a fragment to the debugger.
        This function compiles the fragment to IR, Rel, and SQL, then logs
        the results to the debug system.

        Returns:
            Dictionary containing 'ir', 'rel', and 'sql' keys with the compiled outputs
        """
        try:
            ir_model = self.ir()
            ir_string = ir.node_to_string(ir_model)

            python_source = self.python_source()

            try:
                rel = self.rel(ir_model)
            except Exception:
                rel = None

            try:
                sql = self.sql(ir_model)
            except Exception:
                sql = None

            return CompilationResult(python_source, ir_string, rel, sql)
        except Exception as e:
            raise e

    def display(self, compilation_result):
        """
        Displays the compilation results in a browser by creating a temporary HTML file.

        Args:
            compilation_result: Dictionary containing 'ir', 'rel', and 'sql' keys
                               as returned by the compile() function

        Returns:
            Path to the temporary HTML file
        """
        # Create HTML content with syntax highlighting using highlight.js
        html_content = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>RAI Compilation Results</title>
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/styles/default.min.css">
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/highlight.min.js"></script>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.7.0/languages/sql.min.js"></script>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 20px; }}
                .section {{ margin-bottom: 30px; }}
                h2 {{ color: #333; }}
                pre {{ background-color: #f5f5f5; padding: 10px; border-radius: 5px; overflow: auto; }}
                .error {{ color: red; }}
            </style>
        </head>
        <body>
            <h1>RAI Compilation Results</h1>

            <div class="section">
                <h2>IR (Intermediate Representation)</h2>
                <pre><code class="language-json">{}</code></pre>
            </div>

            <div class="section">
                <h2>Rel</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <div class="section">
                <h2>SQL</h2>
                <pre><code class="language-sql">{}</code></pre>
            </div>

            <script>hljs.highlightAll();</script>
        </body>
        </html>
        """.format(
            html.escape(compilation_result.get('ir', 'Not available')) if compilation_result.get('ir') else '<span class="error">Error during IR compilation</span>',
            html.escape(compilation_result.get('rel', 'Not available')) if compilation_result.get('rel') else '<span class="error">Error during Rel compilation</span>',
            html.escape(compilation_result.get('sql', 'Not available')) if compilation_result.get('sql') else '<span class="error">Error during SQL compilation</span>'
        )

        # Create a temporary HTML file
        with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as f:
            f.write(html_content.encode('utf-8'))
            temp_path = f.name

        # Open the HTML file in the default browser
        webbrowser.open('file://' + os.path.abspath(temp_path))

        return temp_path



from typing import List
from v0.relationalai.early_access.dsl.orm.models import Model, Relationship, Role
from v0.relationalai.early_access.dsl.orm.reasoners import Multiplicity

# Decapitalizes (lowers first character of) a string
def decapitalize(text: str) -> str: return text[:1].lower() + text[1:]

class Verbalizer:
    def __init__(self, model: Model):
        self._model = model

        # We cache incrementally maintainable RelationshipVerbaliers to prevent
        # unnecessary recalculation while users are actively editing a model.
        self._rel_verbs = {}

    def verbalize(self, rel: Relationship) -> str:
        if rel._id not in self._rel_verbs:
            self._rel_verbs[rel._id] = RelationshipVerbalizer(rel, self._model)
        return str(self._rel_verbs[rel._id])


# PointMultiplicityVerbalizers are used to verbalize constraints on the
# Roles of some Relationship using some specific RelationshipReading of
# that Relationship. We verbalize these constraints in the language of
# point multiplicities of a given Role
#
class PointMultiplicityVerbalizer:

    def __init__(self, reading, model: Model):
        self._model = model
        self._reading = reading

    # Use this reading to verbalize the multiplicity of *role* in a context
    # in which players for all of the other roles have been fixed.
    #
    def multiplicity_of(self, role: Role) -> str:
        arity = role._part_of()._arity()
        if arity == 1:
            return self._multiplicity_of_unary(role)
        if arity == 2:
            return self._multiplicity_of_binary(role)
        else:
            return self._multiplicity_of_nary(role)

    def sample_fact(self): return self._reading._sample_fact()

    # Binary relationships admit much more readable multiplicity verbalizations
    # than the more general n-ary relationships; so we handle them separately.
    #
    def _multiplicity_of_binary(self, role: Role):
        reading = self._reading

        role1_text = str(reading._role_in_reading[0])
        pos1_text = str(role1_text) if reading._follows[0] is None else str(role1_text) + " " + reading._follows[0]
        role2_text = str(reading._role_in_reading[1])
        pos2_text = role2_text if reading._follows[1] is None else str(role2_text) + " " + reading._follows[1]

        reasoner = self._model.reasoner()
        mult = reasoner.point_multiplicity(role)
        mult_verb = mult.qualifier()
        leads = reading._leading_text

        if role._id == reading._last()._id:
            # When the role with multiplicity is the second in this reading,
            # the reading is natural
            if mult.many():
                if leads is None:
                    return f"It is possible that some {pos1_text} {mult_verb} {pos2_text}"
                else:
                    return f"It is possible that {decapitalize(leads)} some {pos1_text} {mult_verb} {pos2_text}"
            else:
                if leads is None:
                    return f"Each {pos1_text} {mult_verb} {pos2_text}"
                else:
                    return f"{leads.capitalize()} each {pos1_text} {mult_verb} {pos2_text}"


        else:
            # The role with multiplicity is the first in this reading
            if mult.many():
                if leads is None:
                    return f"It is possible that for some {role2_text}, {mult_verb} {pos1_text} that {pos2_text}"
                else:
                    return f"It is possible that for some {role2_text}, {decapitalize(leads)} {mult_verb} {pos1_text} that {pos2_text}"
            else:
                if leads is None:
                    return f"For each {role2_text}, {mult_verb} {pos1_text} that {pos2_text}"
                else:
                    return f"For each {role2_text}, {decapitalize(leads)} {mult_verb} {pos1_text} that {pos2_text}"

    def _multiplicity_of_unary(self, role: Role) -> str:
        reading = self._reading
        point_variable = str(role.player())
        mult_verb = []
        if reading._leading_text is not None:
            mult_verb.append(decapitalize(reading._leading_text))
        mult_verb.append("that " + str(reading._role_in_reading[0]))
        follows = reading._follows[0]
        if follows is not None:
            mult_verb.append(reading._follows[0])

        return f"It is possible that for some {point_variable}, " + " ".join(mult_verb)

    def _multiplicity_of_nary(self, role: Role) -> str:

        def point_var_seq(pvars: List[str]) -> str:
            if len(pvars) == 1:
                return pvars[0]
            else:
                length = len(pvars) - 1
                return ", ".join(pvars[0:length]) + " and " + pvars[length]

        reading = self._reading

        point_variables = []
        mult_verb = []
        if reading._leading_text is not None:
            mult_verb.append(decapitalize(reading._leading_text))
        idx = 0
        role_verb = None
        for r in reading._roles():
            if role._id == r._id:
                qualifier = Multiplicity.qualifier(self._model.reasoner().point_multiplicity(role))
                role_verb = qualifier + " " + str(reading._role_in_reading[idx])
            else:
                role_verb = "that " + str(reading._role_in_reading[idx])
                point_variables.append(str(r.player()))
            follows = reading._follows[idx]
            if follows is None:
                mult_verb.append(role_verb)
            else:
                mult_verb.append(role_verb + " " + follows)
            idx += 1
    
        return "It is possible that for some " + point_var_seq(point_variables) + ", " + " ".join(mult_verb)


# RelationshipVerbalizers verbalize information about a Relationship in
# natural language, including a sample fact and a verbalization of its
# constraints in the form of a point multiplicity verbalization of each
# of its roles.
# 
# Generating good point multiplicity verbalizations is tricky because:
#   - the most natural way for a human to understand a point multiplicity
#     is when it is verbalized on the last role of some RelationshipReading,
#     but the model may not contain a reading in which a given role appears
#     last.
#   - mandatory and uniqueness constraints on the all of the Roles of the
#     Relationship can influence the point multiplicity of any given Role.
#  
class RelationshipVerbalizer:

    def __init__(self, relationship: Relationship, model: Model):
        self._relationship = relationship
        self._model = model

        # Number of readings of this relationship
        self._number_of_readings = 0   # Forces a sync with the model on first use

        # Maps the last role of a RelationshipReading to the PointMultiplicityVerbalizer
        #   for that reading. Used to simplify the choice of readings to use when
        #   verbalizing the point multiplicity of a Role, as the best ones are when
        #   the role appears last.
        self._role_to_verb = {}

        # Default verbalizer to use when verbalizing a point multiplicity of this
        #   Relationship when there is no better choice. Also used to verbalize
        #   sample facts of this Relationship because it is specific to a Reading.
        self._default_verb = None

    def __str__(self):
        default_reading = self.verbalizer_for()
        verb = [default_reading.sample_fact()]
        for role in self._relationship._rel_roles.values():
            reading_verb = self.verbalizer_for(role)
            verb.append(reading_verb.multiplicity_of(role))
        return "\n".join(verb)

    # Returns a verbalizer that is good for verbalizing the point multiplicity
    # of a given role of this Relationship. If no role is provided, returns
    # a default verbalizer.
    def verbalizer_for(self, role=None) -> PointMultiplicityVerbalizer:
        num_readings = len(self._relationship._readings)
        if self._number_of_readings < num_readings:
            # Then update the reading verbalizations to include those of any
            # newly added RelationshipReadings.
            #
            self._number_of_readings = num_readings
            self._role_to_verb = {}

            for reading in self._relationship._readings:
                reading_verb = PointMultiplicityVerbalizer(reading, self._model)
                self._role_to_verb[reading._last()._id] = reading_verb
                if self._default_verb is None:
                    self._default_verb = reading_verb

        if self._default_verb is None:
            raise Exception(f"Could not find a default reading for Relationship {self._relationship}")

        if role is not None and role._id in self._role_to_verb:
            return self._role_to_verb[role._id]
        
        # Otherwise choose an arbitary reading verbalization
        return self._default_verb


from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field, replace as dc_replace
from typing import  Dict, Sequence as seq, Iterable
from enum import Enum
from math import pi

# PyRel imports
from ..util.naming import sanitize
from ..semantics.metamodel import metamodel as mm
from ..semantics.metamodel.builtins import builtins as b, is_abstract
from ..semantics.metamodel.metamodel_analyzer import Normalize
# ensure std libs are loaded
from ..semantics.metamodel.metamodel_analyzer import VarFinder
from ..semantics.metamodel.rewriter import Walker
from ..semantics.std import aggregates, common, math, numbers, re, strings, decimals, datetime, floats
from ..semantics.backends.lqp import annotations as lqp_annotations

# V0 imports
from v0.relationalai.semantics.metamodel import ir as v0, builtins as v0_builtins, types as v0_types, factory as f
from v0.relationalai.semantics.metamodel.util import FrozenOrderedSet, frozen, ordered_set, filter_by_type, OrderedSet
from v0.relationalai.semantics.internal.internal import literal_value_to_type
from v0.relationalai.semantics.metamodel.typer import typer as v0_typer
from v0.relationalai.clients.util import IdentityParser

from .hoister import Hoister
from .helpers import is_output_update, is_main_output

#------------------------------------------------------
# Simplified translation API
#------------------------------------------------------
def translate(model: mm.Model) -> v0.Model:
    return Translator().translate_model(model)

#------------------------------------------------------
# Implementation
#------------------------------------------------------
class Context(Enum):
    # when translating tasks (i.e. runtime behavior)
    TASK = 1
    # when translating the static structure of the model (types, relations, etc.)
    MODEL = 2

@dataclass
class Translator():

    stack: list[mm.Node] = field(default_factory=list)

    # maps from mm.Node id to the v0.Node created for it
    # for Context=TASK
    task_map: Dict[int, v0.Node] = field(default_factory=dict)
    # for Context=MODEL
    model_map: Dict[int, v0.Node] = field(default_factory=dict)

    # map from a var that is bound to a table, to the list of vars that were used to construct
    # this var. This var is going to be used as a key to an output.
    key_map: Dict[mm.Var, seq[v0.Var]] = field(default_factory=dict)

    # map from a var that is bound to a data, to the data itself. This is used to translate
    # lookups on this var into v0 Data nodes.
    data_map: Dict[mm.Var, mm.Var] = field(default_factory=dict)

    # map from a var that is bound to an external table, to the vars of its fields
    column_map: defaultdict[mm.Var, dict[mm.Relation, v0.Value]] = field(default_factory=lambda: defaultdict(dict))

    # used tables
    used_tables: OrderedSet[mm.Table] = field(default_factory=lambda: ordered_set())

    # when translating a relation that has multiple readings, we need to translate to multiple
    # relations in v0, and we need to add rules to populate those relations. We also need rules
    # to populate super types from sub types.
    maintenance_rules: list[v0.Logical] = field(default_factory=list)

    # outputs that are exports
    exports: set[v0.Output] = field(default_factory=set)

    hoister = Hoister()
    normalize = Normalize()

    def get(self, ctx: Context, node: mm.Node):
        if ctx == Context.MODEL:
            return self.model_map.get(node.id, None)
        else:
            return self.task_map.get(node.id, None)
    def set(self, ctx: Context, node: mm.Node, value: v0.Node):
        if ctx == Context.MODEL:
            self.model_map[node.id] = value  # type: ignore
        else:
            self.task_map[node.id] = value  # type: ignore

    def translate_node(self, node: mm.Node, parent=None, ctx=Context.TASK) -> v0.Node|seq[v0.Node]:
        x = self.get(ctx, node)
        if x is not None:
            return x

        source_class_name = node.__class__.__name__.lower()
        translator = getattr(self, f"translate_{source_class_name}", None)
        if translator is None:
            raise NotImplementedError(f"No translator for node type: {source_class_name}")

        if parent:
            self.stack.append(parent)
        result = translator(node, parent, ctx)
        if parent:
            self.stack.pop()

        self.set(ctx, node, result)
        return result

    def translate_value(self, value: mm.Value, parent, ctx) -> v0.Value:
        # Value = _Union[Var, Literal, Type, Relation, seq["Value"], None]
        if value is None:
            return None
        elif isinstance(value, mm.ScalarType) or isinstance(value, mm.Relation):
            return self.translate_node(value, parent, Context.MODEL) # type: ignore
        elif isinstance(value, seq):
            return self.translate_seq(value, parent, ctx) # type: ignore
        return self.translate_node(value, parent, ctx) # type: ignore


    def translate_seq(self, nodes: seq[mm.Node], parent, ctx) -> tuple[v0.Node,...]:
        res = []
        for n in nodes:
            x = self.translate_node(n, parent, ctx)
            # flat map would be great here
            if isinstance(x, list) or isinstance(x, tuple):
                res.extend(x)
            elif x is not None:
                res.append(x)
        return tuple(res)

    def translate_frozen(self, nodes: seq[mm.Node], parent, ctx) -> FrozenOrderedSet[v0.Node]:
        return frozen(*self.translate_seq(nodes, parent, ctx))

    #------------------------------------------------------
    # Helper
    #------------------------------------------------------

    # NOTE: This has to match what is done in the v0 snowflake.Table class as that is what CDC
    # produces. If there's a quote in the table name, then we take it verbatim, otherwise we lowercase it.
    def translate_table_name(self, table: mm.Table) -> str:
        name = IdentityParser(table.name).identifier
        name = name.lower() if '"' not in table.name else table.name.replace('"', '_')
        return sanitize(name)

    #-----------------------------------------------------------------------------
    # Capabilities, Reasoners
    #-----------------------------------------------------------------------------

    def translate_capability(self, c: mm.Capability, parent: mm.Reasoner, ctx) -> v0.Capability:
        return v0.Capability(name=c.name)

    def translate_reasoner(self, r: mm.Reasoner|None, parent, ctx) -> v0.Engine|None:
        if r is None:
            return None
        return v0.Engine(
            name=f"{r.id}", # no name field in v0, so use id
            platform=r.type,
            info=r.info,
            capabilities=self.translate_frozen(r.capabilities, r, ctx), # type: ignore
            relations=self.translate_frozen(r.relations, r, ctx), # type: ignore
            annotations=self.translate_frozen(r.annotations, r, ctx) # type: ignore
        )

    # -----------------------------------------------------------------------------
    # Relation
    # -----------------------------------------------------------------------------

    def translate_field(self, f: mm.Field, parent, ctx) -> v0.Field:
        return v0.Field(
            name=f.name,
            type=self.translate_node(f.type, f, Context.MODEL), # type: ignore
            input=f.input,
        )

    def translate_reading(self, r: mm.Reading, parent: mm.Relation, ctx) -> v0.Relation:
        # Readings in v0 are represented as separate Relation nodes with different field orders
        fields = []
        for i in r.field_order:
            fields.append(self.translate_field(parent.fields[i], r, ctx))

        # this is not totally correct, we are translating the parent's overloads here, but
        # the order of fields would be different. However, it's unlikely we will have multiple
        # readings AND overloads in the same relation, so this is acceptable for now.
        return v0.Relation(
            name=f"{parent.name}_{parent.readings.index(r)}",
            fields=tuple(fields),
            requires=frozen(),
            annotations=frozen(),
            overloads=self.translate_frozen(parent.overloads, parent, ctx), # type: ignore
        )

    def translate_overload(self, o: mm.Overload, parent: mm.Relation, ctx) -> v0.Relation:
        # Overloads in v0 are represented as separate Relation nodes with different types
        fields = []
        for field, typ in zip(parent.fields, o.types):
            fields.append(
                v0.Field(name=field.name, type=self.translate_node(typ, field, ctx), input=field.input) # type: ignore
            )
        return v0.Relation(
            name=parent.name,
            fields=tuple(fields),
            requires=frozen(),
            annotations=frozen(),
            overloads=frozen()
        )

    # map builtin relations by name
    BUILTIN_RELATION_MAP = {}
    RENAMES = {
        "degrees": "rad2deg",
        "radians": "deg2rad",
        "len": "num_chars",
        "power": "pow",
        "^": "pow",
        "like": "like_match",
        "regex_escape": "escape_regex_metachars",
    }
    for lib in ["core", "aggregates", "common", "datetime", "floats", "math", "numbers", "strings", "re", "lqp"]:
        for x in b._libraries[lib].data.values():
            v0_name = RENAMES.get(x.name, x.name)
            if v0_name in v0_builtins.builtin_relations_by_name:
                BUILTIN_RELATION_MAP[x] = v0_builtins.builtin_relations_by_name[v0_name]

    def translate_relation(self, r: mm.Relation, parent, ctx):
        if r in self.BUILTIN_RELATION_MAP:
            return self.BUILTIN_RELATION_MAP[r]

        if r is b.numbers.parse_number:
            if isinstance(parent, mm.Lookup):
                assert(isinstance(parent.args[1], mm.Var))
                out_type = parent.args[1].type
                assert(isinstance(out_type, mm.NumberType))
                if out_type.scale == 0:
                    return v0_builtins.parse_int128 if out_type.precision > 19 else v0_builtins.parse_int64
                else:
                    return v0_builtins.parse_decimal

        # relations can be accessed by any context; if it exists in the other context, use it
        other_ctx = Context.MODEL if ctx == Context.TASK else Context.TASK
        x = self.get(other_ctx, r)
        if x is not None:
            return x  # type: ignore

        if len(r.readings) > 1:
            # if there are multiple readings, create a relation per reading
            relations = self.translate_seq(r.readings, r, ctx) # type: ignore

            # also create a rule to populate the other readings from the main relation
            main_relation = relations[0]
            assert isinstance(main_relation, v0.Relation)
            main_relation_args = tuple(v0.Var(type=field.type, name=field.name) for field in main_relation.fields)
            for reading in r.readings[1:]:
                relation = relations[r.readings.index(reading)]
                assert isinstance(relation, v0.Relation)
                # lookup from main_relation and derive into this relation
                self.maintenance_rules.append(v0.Logical(
                    engine=None,
                    hoisted=(),
                    body=(
                        v0.Lookup(
                            engine=None,
                            relation=main_relation,
                            args=main_relation_args,
                            annotations=frozen(),
                        ),
                        v0.Update(
                            engine=None,
                            relation=relation,
                            args=tuple([main_relation_args[idx] for idx in reading.field_order]),
                            effect=v0.Effect.derive,
                            annotations=frozen(),
                        )
                    ),
                    annotations=frozen(),
                ))
            return relations
        else:
            # if there are no readings or a single reading, create a single relation

            # if this is a placeholder relation, fill overloads with all non-placeholder relations
            # with the same name
            overloads = ordered_set()
            if all(field.type == b.core.Any for field in r.fields):
                for rr in self.relations_by_name[r.name]:
                    if any(field.type != b.core.Any for field in rr.fields):
                        x = self.translate_relation(rr, r, ctx)
                        if isinstance(x, v0.Relation):
                            overloads.add(x)
                        else:
                            overloads.update(x) # type: ignore
            overloads.update(self.translate_frozen(r.overloads, r, ctx))  # type: ignore
            annotations = self.translate_seq(r.annotations, r, ctx)  # type: ignore
            name = r.name
            fields = self.translate_seq(r.fields, r, ctx)  # type: ignore
            # We need to turn column relations into what the CDC pass would otherwise produce
            # by making the relation name be the table name and adding a symbol field at the front
            # representing the column name
            if r.fields and isinstance(r.fields[0].type, mm.Table) and r in r.fields[0].type.columns:
                name = self.translate_table_name(r.fields[0].type)
                fields = (v0.Field("symbol", v0_types.Symbol, False), *fields) # type: ignore
                annotations = annotations + (v0_builtins.external_annotation,)

            return v0.Relation(
                name=name,
                fields=frozen(*fields), # type: ignore
                requires=self.translate_frozen(r.requires, r, ctx), # type: ignore
                annotations=frozen(*annotations), # type: ignore
                overloads=overloads.frozen(), # type: ignore
            )

    def translate_unresolvedrelation(self, r: mm.UnresolvedRelation, parent, ctx):
        # treat unresolved relations the same as normal relations for now
        return self.translate_relation(r, parent, ctx)

    #------------------------------------------------------
    # Annotation
    #------------------------------------------------------

    def translate_annotation(self, a: mm.Annotation, parent, ctx) -> v0.Annotation:
        if a.relation.name in v0_builtins.builtin_annotations_by_name:
            return getattr(v0_builtins, a.relation.name + "_annotation")  # type: ignore

        return v0.Annotation(
            relation=self.translate_node(a.relation, a, Context.MODEL), # type: ignore
            args=tuple(self.translate_value(arg, a, ctx) for arg in a.args)
        )

    #-----------------------------------------------------------------------------
    # Types
    #-----------------------------------------------------------------------------

    BUILTIN_TYPE_MAP = {
        # Abstract types (should be removed once our typer is done because they should not
        # show up in the typed metamodel.)
        b.core.Any: v0_types.Any,
        # b.core.AnyEntity: v0_types.AnyEntity,
        b.core.Number: v0_types.Number, # v0 typer can figure the rest out
        # Concrete types
        b.core.Boolean: v0_types.Bool,
        b.core.String: v0_types.String,
        b.core.Date: v0_types.Date,
        b.core.DateTime: v0_types.DateTime,
        b.core.Float: v0_types.Float,
        b.core.Hash: v0_types.Hash,
    }

    def translate_scalartype(self, t: mm.ScalarType, parent: mm.Node, ctx) -> v0.ScalarType|v0.Relation|None:
        if t in self.BUILTIN_TYPE_MAP:
            return self.BUILTIN_TYPE_MAP[t]

        if ctx == Context.TASK:
            # in task context, this is a lookup or update on a concept population relation
            # TODO - removing this filter for now
            actual_type = self.translate_node(t, t, Context.MODEL)
            assert isinstance(actual_type, v0.ScalarType)
            fields = [v0.Field(name="entity", type=actual_type, input=False)] # type: ignore
            annotations = [v0_builtins.concept_relation_annotation]
            name = t.name
            if isinstance(t, mm.Table):
                annotations.append(v0_builtins.external_annotation)
                name = self.translate_table_name(t)
                fields.insert(0, v0.Field(name="symbol", type=v0_types.Symbol, input=False)) # type: ignore

            type_relation = v0.Relation(
                name=name,
                fields=tuple(fields), # type: ignore
                requires=frozen(),
                annotations=frozen(*annotations), # type: ignore
                overloads=frozen()
            )
            for super_type in t.super_types:
                super_rel = self.translate_node(super_type, t, Context.TASK)
                if not super_rel:
                    continue
                assert isinstance(super_rel, v0.Relation)
                v = v0.Var(type=actual_type, name=type_relation.name.lower())
                self.maintenance_rules.append(v0.Logical(
                    engine=None,
                    hoisted=(),
                    body=(
                        v0.Lookup(
                            engine=None,
                            relation=type_relation,
                            args=(v,),
                            annotations=frozen(),
                        ),
                        v0.Update(
                            engine=None,
                            relation=super_rel,
                            args=(v,),
                            effect=v0.Effect.derive,
                            annotations=frozen(),
                        )
                    ),
                    annotations=frozen(),
                ))
            return type_relation

        extends=[]
        if isinstance(t, mm.Table):
            extends.append(v0_types.RowId)

        for st in t.super_types:
            # for now make all value types extending number be Int64, we will fix this with a new typer
            if st == b.core.Number:
                translated_st = v0_types.Number
            else:
                translated_st = self.translate_node(st, t, ctx)
            extends.append(translated_st)

        return v0.ScalarType(
            name=t.name,
            super_types=frozen(*extends), # type: ignore
            # super_types=self.translate_frozen(t.super_types, t, ctx), # type: ignore
            annotations=self.translate_frozen(t.annotations, t, ctx) # type: ignore
        )

    def translate_numbertype(self, t: mm.NumberType, parent, ctx):
        if t.scale == 0:
            return v0_types.Int128 if t.precision > 19 else v0_types.Int64
        return v0.DecimalType(
            name=t.name,
            precision=t.precision,
            scale=t.scale,
            super_types=frozen(),
            annotations=self.translate_frozen(t.annotations, t, ctx) # type: ignore
        )

    def translate_uniontype(self, t: mm.UnionType, parent, ctx) -> v0.UnionType:
        return v0.UnionType(
            types=self.translate_frozen(t.types, t, ctx), # type: ignore
        )

    def translate_listtype(self, t: mm.ListType, parent, ctx) -> v0.ListType:
        return v0.ListType(
            element_type=self.translate_node(t.element_type, t, ctx), # type: ignore
        )

    def translate_tupletype(self, t: mm.TupleType, parent, ctx) -> v0.TupleType:
        return v0.TupleType(
            types=self.translate_seq(t.element_types, t, ctx), # type: ignore
        )


    #------------------------------------------------------
    # Table
    #------------------------------------------------------

    # TODO - this depends on what is using it, it may become an Output
    # class Table(ScalarType):
    #     columns: seq[Relation] = tuple_field()
    #     uri: str = ""

    def translate_table(self, t: mm.Table, parent, ctx):
        return self.translate_scalartype(t, parent, ctx)

    def rewrite_cdc_args(self, l: mm.Lookup, args, parent, ctx):
        # If this is a lookup for an external table column,
        # we have to prepend the column symbol to the args
        root_type = l.relation.fields[0].type
        if isinstance(root_type, mm.Table):
            self.used_tables.add(root_type)
            is_table = l.relation == root_type
            if is_table or l.relation in root_type.columns:
                sym = "METADATA$KEY" if is_table else l.relation.name
                args = (v0.Literal(type=v0_types.Symbol, value=sym), *args)
        return args

    # -----------------------------------------------------------------------------
    # Values
    # -----------------------------------------------------------------------------

    # Primitive = _Union[str, int, float, bool, _dt.datetime, None]

    def translate_var(self, v: mm.Var, parent, ctx) -> v0.Var:
        if isinstance(parent, mm.Lookup):
            root_type = parent.relation.fields[0].type
            root_arg = parent.args[0]
            if isinstance(root_arg, mm.Var) \
                and isinstance(root_type, mm.Table) \
                and (existing := self.column_map[root_arg].get(parent.relation)):
                return existing # type: ignore

        return v0.Var(
            type=self.translate_node(v.type, v, Context.MODEL), # type: ignore
            name=v.name
        )

    def translate_literal(self, l: mm.Literal, parent, ctx) -> v0.Literal:
        typ = None
        if l.type is None or is_abstract(l.type):
            # force int64 for literals, it can be widened later if needed
            if  type(l.value) == int:
                typ = v0_types.Int64
            else:
                # using v0's map of literal to type
                typ = literal_value_to_type(l.value)
        else:
            typ = self.translate_node(l.type, l, Context.MODEL)
        if type is None:
            pass
        return v0.Literal(
            type=typ, # type: ignore
            value=l.value
        )

    # Value = _Union[Var, Literal, Type, Relation, seq["Value"], None]

    # # -----------------------------------------------------------------------------
    # # Tasks (process algebra)
    # # -----------------------------------------------------------------------------

    # @dataclass(frozen=True, kw_only=True)
    # class Task(Node):
    #     reasoner: Optional[Reasoner] = None

    # -----------------------------
    # Control Flow
    # -----------------------------

    def _merge_outputs(self, children: seq[v0.Node], outputs:list[v0.Output]):
        # We need to always rewrite the outputs, even if there's only one, because
        # we may need to mark it as an export and handle the NO_KEYS case.
        # if len(outputs) < 2:
        #     return children

        children = list(children)

        # merge all outputs, accumulating keys and aliases
        keys = ordered_set()
        aliases = ordered_set()
        for output in outputs:
            assert isinstance(output, v0.Output)
            children.remove(output)
            if output.keys is not None:
                keys.update(output.keys)
            aliases.update(output.aliases)

        # add a single, merged output
        is_export = any(output in self.exports for output in outputs)
        annos = (v0_builtins.export_annotation,) if is_export else ()
        if is_export and not keys:
            key_var = v0.Var(type=v0_types.Hash, name="export_key")
            vals = [alias[1] for alias in aliases]
            children.append(v0.Construct(
                engine=None,
                values=(v0.Literal(type=v0_types.String, value="NO_KEYS"), *vals),
                id_var=key_var,
            ))
            keys.add(key_var)

        children.append(
                v0.Output(
                    engine=None,
                    aliases=frozen(*aliases),
                    keys=frozen(*keys),
                    annotations=frozen(*annos)
                )
            )
        return tuple(children)

    def get_outputs(self, children: seq[v0.Node]) -> list[v0.Output]:
        return list(filter(lambda o: isinstance(o, v0.Output), children)) # type: ignore

    def translate_logical(self, l: mm.Logical, parent, ctx):
        # first process children
        children = self.translate_seq(l.body, l, ctx)

        # inline logicals if possible
        new_children = []
        for c in children:
            if isinstance(c, v0.Logical) and not c.hoisted and len(c.body) == 1 and not isinstance(c.body[0], (v0.Aggregate, v0.Rank)):
                new_children.extend(c.body)
            else:
                new_children.append(c)
        # children = children if len(children) == len(new_children) else tuple(new_children)
        children = tuple(new_children)

        # compute variables to hoist, wrapping in Default if the logical is optional
        if l.optional:
            hoisted = tuple([v0.Default(self.translate_node(v, l, ctx), None) for v in self.hoister.hoisted(l)]) # type: ignore
        else:
            hoisted = tuple([self.translate_node(v, l, ctx) for v in self.hoister.hoisted(l)]) # type: ignore

        # Nots can never hoist variables upwards
        if isinstance(parent, mm.Not):
            hoisted = ()

        # if there are no outputs, just return the logical
        outputs = self.get_outputs(children)
        if not outputs:
            return v0.Logical(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                hoisted=hoisted, # type: ignore
                body=children, # type: ignore
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )


        # if there's a main output here, we need to merge all outputs as a single one
        has_main_output = any(is_main_output(c) for c in l.body)
        if has_main_output:
            return v0.Logical(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                hoisted=hoisted, # type: ignore
                body=self._merge_outputs(children, outputs), # type: ignore
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )
        else:
            # this logical has an output but it's not the main one; so we return the outputs
            # side-by-side with the logical. This assumes that the parent is a logical that
            # will splat these tasks and we will eventually get into a logical with a main
            # output that will merge them all.

            # remove outputs from children
            children = tuple([c for c in children if not isinstance(c, v0.Output)])
            if not children:
                return outputs

            # if this is an optional logical but we're not hoisting anything and not updating,
            # then it's possible we're filtering an outer variable, but only for this column. If
            # so, we need to alias the output and hoist it.
            # this is important because the LQP stack blows up if there's a logical with no effect
            if l.optional and not hoisted and not any(isinstance(c, v0.Update) for c in children):
                # if there are no lookups, then this really is a no-op, just return outputs
                # LQP blows up with e.g. a match-only logical
                if not any(isinstance(c, v0.Lookup) for c in children):
                    return outputs
                # otherwise, make sure we filter the outer variable through aliasing
                new_children = [*children]
                new_hoists = []
                new_outputs = []
                # add an eq to a new var for the output, hoist the new var, change the output to use the new var
                for output in outputs:
                    # shim outputs always only have one alias
                    (name, orig_var) = output.aliases.data[0]
                    # if the original var is not a Var or Literal, or it's already in the keys, skip (we can't filter keys)
                    if not isinstance(orig_var, (v0.Var, v0.Literal)) or (output.keys and orig_var in output.keys): #type: ignore
                        new_outputs.append(output)
                        continue
                    new_var = v0.Var(
                        type=orig_var.type,
                        name=f"{orig_var.name if isinstance(orig_var, v0.Var) else 'literal'}_hoisted"
                    )
                    new_output = v0.Output(
                        engine=None,
                        aliases=frozen((name, new_var)),
                        keys=output.keys,
                        annotations=output.annotations
                    )
                    eq = v0.Lookup(
                        engine=None,
                        relation=v0_builtins.eq,
                        args=(new_var, orig_var),
                        annotations=frozen()
                    )
                    new_children.append(eq)
                    new_hoists.append(v0.Default(new_var, None))
                    new_outputs.append(new_output)
                outputs = new_outputs
                hoisted = tuple(new_hoists)
                children = tuple(new_children)
                # return outputs

            # return outputs + a logical with the other children
            outputs.append(
                v0.Logical(
                    engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                    hoisted=hoisted, # type: ignore
                    body=children, # type: ignore
                    annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
                ))
            return outputs




    def translate_sequence(self, s: mm.Sequence, parent, ctx) -> v0.Sequence:
        return v0.Sequence(
            engine=self.translate_reasoner(s.reasoner, s, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(s), s, ctx), # type: ignore
            tasks=self.translate_seq(s.tasks, s, ctx), # type: ignore
            annotations=self.translate_frozen(s.annotations, s, ctx) # type: ignore
        )

    def translate_union(self, u: mm.Union, parent, ctx) -> v0.Union:
        return v0.Union(
            engine=self.translate_reasoner(u.reasoner, u, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(u), u, ctx), # type: ignore
            tasks=self.translate_seq(u.tasks, u, ctx), # type: ignore
            annotations=self.translate_frozen(u.annotations, u, ctx) # type: ignore
        )

    def translate_match(self, m: mm.Match, parent, ctx) -> v0.Match:
        return v0.Match(
            engine=self.translate_reasoner(m.reasoner, m, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(m), m, ctx), # type: ignore
            tasks=self.translate_seq(m.tasks, m, ctx), # type: ignore
            annotations=self.translate_frozen(m.annotations, m, ctx) # type: ignore
        )

    def translate_until(self, u: mm.Until, parent, ctx) -> v0.Until:
        return v0.Until(
            engine=self.translate_reasoner(u.reasoner, u, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(u), u, ctx), # type: ignore
            check=self.translate_node(u.check, u, ctx), # type: ignore
            body=self.translate_node(u.body, u, ctx), # type: ignore
            annotations=self.translate_frozen(u.annotations, u, ctx) # type: ignore
        )

    def translate_wait(self, w: mm.Wait, parent, ctx) -> v0.Wait:
        return v0.Wait(
            engine=self.translate_reasoner(w.reasoner, w, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(w), w, ctx), # type: ignore
            check=self.translate_node(w.check, w, ctx), # type: ignore
            annotations=self.translate_frozen(w.annotations, w, ctx) # type: ignore
        )

    def translate_loop(self, l: mm.Loop, parent, ctx) -> v0.Loop:
        # TODO - loop is incompatible because over has multiple vars, so this is a best
        # attempt (does not matter much as this is not used in practice)
        return v0.Loop(
            engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
            hoisted=self.translate_seq(self.hoister.hoisted(l), l, ctx), # type: ignore
            iter=self.translate_node(l.over[0], l, ctx), # type: ignore
            body=self.translate_node(l.body, l, ctx), # type: ignore
            annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
        )

    def translate_break(self, b: mm.Break, parent, ctx) -> v0.Break:
        return v0.Break(
            engine=self.translate_reasoner(b.reasoner, b, Context.MODEL),
            check=self.translate_node(b.check, b, ctx), # type: ignore
            annotations=self.translate_frozen(b.annotations, b, ctx) # type: ignore
        )

    # -----------------------------
    # Constraints
    # -----------------------------

    def translate_require(self, r: mm.Require, parent, ctx) -> v0.Require|None:
        # check if this is a unique_fields constraint
        assert isinstance(r.domain, mm.Logical)
        assert isinstance(r.check, mm.Logical)
        if not r.domain.body and all(isinstance(c, mm.Lookup) and c.relation is b.constraints.unique_fields for c in r.check.body):
            return
            v0_reqs = []
            # v0 expects a check with both the domain and the relation in it followed by the unique
            # constraint, so we construct a logical with all of that in it
            for c in r.check.body:
                assert isinstance(c, mm.Lookup)
                fields = c.args[0] # tuple of fields
                assert isinstance(fields, tuple) and all(isinstance(f, mm.Field) for f in fields)
                first_field = fields[0]
                if isinstance(first_field.type, mm.Table):
                    # skip union types for now
                    continue
                assert isinstance(first_field, mm.Field)
                relation = first_field._relation
                domain = self.translate_node(mm.Logical(()), r, ctx)
                all_vars =  [mm.Var(type=field.type, name=field.name + "_var") for field in relation.fields]
                check_tasks = [
                    mm.Lookup(relation, tuple(all_vars)),
                    mm.Lookup(b.constraints.unique, (tuple(all_vars[0:len(fields)]),))
                ]
                check_logical = self.translate_node(mm.Logical(tuple(check_tasks)), r, Context.TASK)
                check = v0.Check(
                    check=check_logical, # type: ignore
                    error=None,
                    annotations=self.translate_frozen(r.annotations, r, ctx) # type: ignore
                )
                v0_reqs.append(v0.Require(
                    engine=self.translate_reasoner(r.reasoner, r, Context.MODEL),
                    domain=domain, # type: ignore
                    checks=(check,), # type: ignore
                    annotations=self.translate_frozen(r.annotations, r, ctx) # type: ignore
                ))
            return v0_reqs # type: ignore

        # check = v0.Check(
        #     check=self.translate_node(r.check, r, ctx), # type: ignore
        #     error=self.translate_node(r.error, r, ctx) if r.error else None, # type: ignore
        #     annotations=self.translate_frozen(r.annotations, r, ctx) # type: ignore
        # )
        # return v0.Require(
        #     engine=self.translate_reasoner(r.reasoner, r, Context.MODEL),
        #     domain=self.translate_node(r.domain, r, ctx), # type: ignore
        #     checks=(check,), # type: ignore
        #     annotations=self.translate_frozen(r.annotations, r, ctx) # type: ignore
        # )
        return None

    # -----------------------------
    # Quantifiers
    # -----------------------------

    def translate_not(self, n: mm.Not, parent, ctx) -> v0.Not:
        return v0.Not(
            engine=self.translate_reasoner(n.reasoner, n, Context.MODEL),
            task=self.translate_node(n.task, n, ctx), # type: ignore
            annotations=self.translate_frozen(n.annotations, n, ctx) # type: ignore
        )

    def translate_exists(self, e: mm.Exists, parent, ctx) -> v0.Exists:
        return v0.Exists(
            engine=self.translate_reasoner(e.reasoner, e, Context.MODEL),
            vars=self.translate_seq(e.vars, e, ctx), # type: ignore
            task=self.translate_node(e.task, e, ctx), # type: ignore
            annotations=self.translate_frozen(e.annotations, e, ctx) # type: ignore
        )

    # -----------------------------
    # Relation Ops
    # -----------------------------

    def translate_effect(self, e: mm.Effect, parent, ctx) -> v0.Effect:
        if e == mm.Effect.derive:
            return v0.Effect.derive
        elif e == mm.Effect.insert:
            return v0.Effect.insert
        elif e == mm.Effect.delete:
            return v0.Effect.delete

    def translate_lookup(self, l: mm.Lookup, parent, ctx):
        # Data Node lookups
        if isinstance(l.relation, mm.Data):
            vars = [self.translate_value(l.args[0], l, ctx)]
            for col in l.relation.columns:
                v = v0.Var(
                    type=self.translate_node(col.fields[-1].type, col, Context.MODEL), # type: ignore
                    name=f"{col.name}"
                )
                vars.append(v)
                assert isinstance(l.args[0], mm.Var)
                self.column_map[l.args[0]][col] = v
            return v0.Data(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                data=l.relation.data,
                vars=frozen(*vars) # type: ignore
            )
        elif l.args and l.args[0] in self.column_map and l.relation in self.column_map[l.args[0]]: # type: ignore
            assert isinstance(l.args[0], mm.Var)
            var = self.translate_value(l.args[1], l, ctx)
            col_var = self.column_map[l.args[0]][l.relation]  # ensure entry exists
            if var != col_var:
                return v0.Lookup(
                    engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                    relation=v0_builtins.eq,
                    args=(var, col_var), # type: ignore
                    annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
                )
            # if this is a data column, we just ignore it as the data node already binds the variables
            elif isinstance(l.args[0].type, mm.Data):
                return None
            # Otherwise we keep the lookup because that's what the LQP stack expect (the lookup gets repeated for
            # each column)


        relation, args = self._resolve_reading(l, ctx)
        if relation is None:
            return None


        # Specific rewrites
        rewrite = self.rewrite_lookup(l, parent, ctx)
        if rewrite is None:
            args = self.rewrite_cdc_args(l, args, parent, ctx)
            # General translation
            rewrite = v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                relation=relation, # type: ignore
                args=args,
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )

        return rewrite


    def translate_update(self, u: mm.Update, parent, ctx) -> v0.Update|v0.Output|None:
        if isinstance(u.relation, mm.Table):
            # ignore updates to tables for now
            return None

        if is_output_update(u):
            # this is an output, the last arg is the value for the output column, the other
            # args are part of the key

            # get the list of key vars created during construction
            keys = list(self.key_map[u.args[0]]) # type: ignore
            # if there are no keys in the key map, this is a constructor of output without
            # keys; we still need to remove the output from the args, so count is 1
            arg_count = len(keys) if keys else 1
            # append any other args used for the key
            for arg in u.args[arg_count:-1]:
                x = self.translate_value(arg, u, ctx)
                keys.append(x) # type: ignore
            if any(anno.relation is b.core["output_value_is_key"] for anno in u.annotations):
                # if the output_value_is_key annotation is present, add the output value to the keys
                keys.append(self.translate_value(u.args[-1], u, ctx)) # type: ignore

            # there will be a single alias here, the output column
            aliases = ordered_set((u.relation.name, self.translate_value(u.args[-1], u, ctx))).frozen()

            out = v0.Output(
                engine=self.translate_reasoner(u.reasoner, u, Context.MODEL),
                keys=frozen(*keys), # type: ignore
                aliases=aliases,
                annotations=self.translate_frozen(u.annotations, u, ctx) # type: ignore
            )
            table = u.relation.fields[0].type
            if isinstance(table, mm.Table) and not table.uri.startswith("dataframe://"):
                self.exports.add(out)
            return out
        else:
            relation, args = self._resolve_reading(u, ctx)

            return v0.Update(
                engine=self.translate_reasoner(u.reasoner, u, Context.MODEL),
                relation=relation, # type: ignore
                args=args,
                effect=self.translate_effect(u.effect, u, ctx),
                annotations=self.translate_frozen(u.annotations, u, ctx) # type: ignore
            )

    def _resolve_reading(self, task: mm.Lookup|mm.Update, ctx):
        relation = self.translate_node(task.relation, task, ctx)
        args = tuple(self.translate_value(arg, task, ctx) for arg in task.args)
        if isinstance(relation, seq):
            # Updates should always write to the root so that the reading maintenance rules
            # can populate the other readings
            if task.reading_hint is not None and isinstance(task, mm.Lookup):
                relation = relation[task.relation.readings.index(task.reading_hint)]
                args = tuple([args[i] for i in task.reading_hint.field_order])
            else:
                relation = relation[0]

        return relation, args


    def translate_aggregate(self, a: mm.Aggregate, parent, ctx) -> v0.Aggregate:

        # hoist the return variables (those that are not inputs to the aggregation)
        return_args = []
        for field, arg in zip(a.aggregation.fields, a.args):
            if not field.input:
                return_args.append(arg)
        hoisted=self.translate_seq(return_args, a, ctx)

        # body must contain the aggregate body plus the aggregate itself
        body = self.translate_node(a.body, a, ctx)
        body = list(body.body if isinstance(body, v0.Logical) else [body])

        # construct the aggregate node
        engine = self.translate_reasoner(a.reasoner, a, Context.MODEL)
        aggregation = self.translate_node(a.aggregation, a, ctx)
        projection = self.translate_seq(a.projection, a, ctx)
        group = self.translate_seq(a.group, a, ctx)
        args = tuple(self.translate_value(arg, a, ctx) for arg in a.args)
        annotations = self.translate_frozen(a.annotations, a, ctx)

        v0_agg = None
        if a.aggregation == b.aggregates.rank:
            assert isinstance(args[0], tuple)
            assert isinstance(args[1], tuple)
            v0_agg = v0.Rank(engine, projection, group, args[0], tuple(l.value for l in args[1]), args[-1], 0, annotations,) # type: ignore
        elif a.aggregation == b.aggregates.limit:
            assert isinstance(args[0], v0.Literal)
            assert isinstance(args[1], tuple)
            assert isinstance(args[2], tuple)
            result = v0.Var(type=v0_types.Int64, name="limit_result")
            v0_agg = v0.Rank(engine, projection, group, args[1], tuple(l.value for l in args[2]), result, args[0].value, annotations,) # type: ignore
        else:
            v0_agg = v0.Aggregate(engine, aggregation, projection, group, args, annotations) # type: ignore

        body.append(v0_agg)
        return v0.Logical(
            engine=None,
            hoisted=hoisted, # type: ignore
            body=tuple(body) # type: ignore
        )

    def translate_construct(self, c: mm.Construct, parent, ctx) -> v0.Construct|None:
        if isinstance(c.id_var.type, mm.Table):
            # we are constructing the key for an output. If there are no values, there is no
            # key in v0, so we can just ignore this
            if not c.values:
                self.key_map[c.id_var] = []
                return None
            # now create a new v0 var to represent the hashed key, record that this key is
            # to be used by updates, and generate the constructor
            v = v0.Var(type=v0_types.Hash, name=c.id_var.name)
            self.key_map[c.id_var] = [v]
            values = [v0_types.Hash]
            values.extend(self.translate_seq(c.values, c, ctx)) # type: ignore
            return v0.Construct(
                engine=self.translate_reasoner(c.reasoner, c, Context.MODEL),
                values=tuple(values), # type: ignore
                id_var=v,
                annotations=self.translate_frozen(c.annotations, c, ctx) # type: ignore
            )
        return v0.Construct(
            engine=self.translate_reasoner(c.reasoner, c, Context.MODEL),
            values=self.translate_seq(c.values, c, ctx), # type: ignore
            id_var=self.translate_node(c.id_var, c, ctx), # type: ignore
            annotations=self.translate_frozen(c.annotations, c, ctx) # type: ignore
        )


    def translate_data(self, d: mm.Data, parent, ctx):
        if isinstance(parent, mm.Field) or isinstance(parent, mm.Var):
            return v0_types.Int128

    # -----------------------------------------------------------------------------
    # Model
    # -----------------------------------------------------------------------------

    def translate_model(self, source_model: mm.Model, parent=None, ctx=None) -> v0.Model:
        try:
            # first analyze the model to compute variables that need to be hoisted
            self.hoister.analyze(source_model)

            self.relations_by_name = defaultdict(list)
            for r in source_model.relations:
                self.relations_by_name[r.name].append(r)

            # translate root first as it may create relations from types (e.g. concept populations)
            root = self.translate_node(source_model.root, source_model, Context.TASK)

            # translate all declared relations
            relations:OrderedSet[v0.Relation] = ordered_set(*self.translate_seq(source_model.relations, source_model, Context.MODEL)) # type: ignore
            # add relations created during task translation
            relations.update(filter_by_type(self.task_map.values(), v0.Relation)) # type: ignore
            # v0 needs all overloads also added to the model
            for r in list(relations):
                relations.update(r.overloads) # type: ignore

            # if there are reading maintenance rules, add them as well
            if self.maintenance_rules:
                root = v0.Logical(
                    engine=None,
                    hoisted=(),
                    body=(
                        root, # type: ignore
                        *self.maintenance_rules
                    ),
                    annotations=frozen()
                )
            engines=self.translate_frozen(source_model.reasoners, source_model, Context.MODEL) # type: ignore
            types=self.translate_frozen(source_model.types, source_model, Context.MODEL) # type: ignore

            return v0.Model(
                engines=engines, # type: ignore
                relations=relations.frozen(), # type: ignore
                types=types, # type: ignore
                root=root # type: ignore
            )
        finally:
            self.hoister.clear()

    # -----------------------------------------------------------------------------
    # Translate query
    # -----------------------------------------------------------------------------

    def translate_query(self, source_query: mm.Task) -> v0.Logical:
        try:
            # first analyze the model to compute variables that need to be hoisted
            self.hoister.analyze_task(source_query)

            # translate the query logical
            return self.translate_node(source_query, source_query, Context.TASK) # type: ignore
        finally:
            self.hoister.clear()

    # -----------------------------------------------------------------------------
    # Library-specific rewrites
    # -----------------------------------------------------------------------------

    def adjust_index(self, l: mm.Lookup, indices: list[int], ctx, args=None, no_outputs=False):
        """ Rewrite the lookup such that the args at `indices` are adjusted to convert from
        0-based indexing (as in the mm) to 1-based indexing (as in v0's backend).

        If the field at the given index is input, we increment the arg by 1 before passing
        it to the lookup. If the field is output, we create a tmp var to hold the result
        of the lookup, and then create a subsequent task to decrement the tmp var by 1.

        `args` can be used to provide a different set of args than l.args, in case they need
        reordering (e.g. strings.split).
        """
        if args is None:
            args = l.args
        # vars to hold the result from the v0 lookup
        tmps = {}
        for i in indices:
            index_type = l.relation.fields[i].type
            tmps[i] = mm.Var(type=index_type, name=f"tmp")
        # insert tmp at index
        replaced_args = []
        for i, arg in enumerate(args):
            if i in indices:
                replaced_args.append(tmps[i])
            else:
                replaced_args.append(arg)
        # translate the lookup
        args = tuple(self.translate_value(arg, l, ctx) for arg in replaced_args)
        args = self.rewrite_cdc_args(l, args, l, ctx)
        lookup = v0.Lookup(
            engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
            relation=self.translate_node(l.relation), # type: ignore
            args=args,
            annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
        )
        # subtract 1 from the index to convert from 1-based to 0-based
        tasks = []
        for index in indices:
            arg = args[index]
            if isinstance(arg, mm.Literal):
                if l.relation.fields[index].input:
                    new = mm.Literal(arg.type, arg.value + 1) # type: ignore
                else:
                    new = mm.Literal(arg.type, arg.value + 1) # type: ignore
                tasks.append(v0.Lookup(
                        engine=None,
                        relation=v0_builtins.eq, # type: ignore
                        args=(
                            self.translate_value(tmps[index], lookup, ctx),
                            self.translate_value(new, lookup, ctx),
                        )
                ))
            elif l.relation.fields[index].input:
                tasks.append(v0.Lookup(
                        engine=None,
                        relation=v0_builtins.plus, # type: ignore
                        args=(
                            self.translate_value(args[index], lookup, ctx),
                            v0.Literal(v0_types.Int128, 1),
                            self.translate_value(tmps[index], lookup, ctx),
                        ),
                    )
                )
            else:
                tasks.append(v0.Lookup(
                        engine=None,
                        relation=v0_builtins.minus, # type: ignore
                        args=(
                            self.translate_value(tmps[index], lookup, ctx),
                            v0.Literal(v0_types.Int128, 1),
                            self.translate_value(args[index], lookup, ctx),
                        ),
                    )
                )
        # do casts
        tasks.extend(self.cast(lookup, ctx, no_outputs=no_outputs))

        return tasks

    def decrement(self, l: mm.Lookup, index: int, ctx):
        """ Rewrite the lookup such that the arg at `index` is decremented by 1 before the
        lookup.  """
        x = l.args[index]
        if isinstance(x, mm.Literal):
            # if the arg is a literal, just decrement the literal directly
            new_literal = mm.Literal(type=x.type, value=x.value - 1) # type: ignore
            return v0.Lookup(
                    engine=None,
                    relation=self.translate_node(l.relation), # type: ignore
                    args=tuple(self.translate_value(arg, l, ctx) if i != index else self.translate_value(new_literal, l, ctx) for i, arg in enumerate(l.args)),
                    annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )

        # arg is not a literal, so we need to create a tmp var to store the decremented value
        tmp = self.translate_value(mm.Var(type=b.core.Number, name="tmp"), l, ctx)
        # lookup(..., tmp, ...)
        new = v0.Lookup(
                engine=None,
                relation=self.translate_node(l.relation), # type: ignore
                args=tuple(self.translate_value(arg, l, ctx) if i != index else tmp for i, arg in enumerate(l.args)),
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
        )
        casted = self.cast(new, ctx)
        return (
            # tmp = l.args[index] - 1
            v0.Lookup(
                engine=None,
                relation=v0_builtins.minus, # type: ignore
                args=(
                    self.translate_value(l.args[index], l, ctx),
                    v0.Literal(v0_types.Int128, 1),
                    tmp,
                ),
            ),
            *casted
        )

    def cast(self, l: v0.Lookup, ctx, no_outputs=False):
        # note that everything here is in v0
        inputs = []
        outputs = []
        args = []
        fields = l.relation.fields
        # if there are overloads, we need to cast based on the most compatible overload
        if l.relation.overloads:
            inf = float("inf")
            min_cost = inf
            for overload in l.relation.overloads:
                total = 0
                for arg, field in zip(l.args, overload.fields):
                    if v0_typer.to_type(arg) != field.type:
                        total += 1
                if total < min_cost:
                    min_cost = total
                    fields = overload.fields

        for arg, field in zip(l.args, fields):
            target_type = field.type
            if target_type is None or not isinstance(arg, (v0.Var, v0.Literal)) or arg.type == target_type:
                args.append(arg)
                continue
            if field.input:
                cast_var = v0.Var(type=target_type, name="cast_var")
                args.append(cast_var)
                inputs.append(v0.Lookup(
                    engine=None,
                    relation=v0_builtins.cast, # type: ignore
                    args=(target_type, arg, cast_var),
                ))
            elif not no_outputs:
                cast_var = v0.Var(type=target_type, name="cast_var")
                args.append(cast_var)
                outputs.append(v0.Lookup(
                    engine=None,
                    relation=v0_builtins.cast, # type: ignore
                    args=(arg.type, cast_var, arg),
                ))
            else:
                # This is ridiculous, but there are cases where we need to force the type
                # of the var, but not through an actual cast operation. E.g. due to the way
                # the LQP stack handles dates.
                object.__setattr__(arg, 'type', target_type)
                args.append(arg)

        lookup = v0.Lookup(
            engine=l.engine,
            relation=l.relation,
            args=tuple(args),
            annotations=l.annotations
        )
        if no_outputs:
            outputs = []
        return [*inputs, lookup, *outputs]

    def rewrite_lookup(self, l: mm.Lookup, parent, ctx):
        """ Special cases for v1 builtins that either have some different shape in v0 or
        do not exist as a primitive in the engine, so we want to compose the result so that
        LQP works. Return None if no rewrite is needed.

        Note that this type of adjustment will need to be made by emitters anyways, so we
        will want to have a good API for them. But in that case, they should do
        transformations still in the mm, whereas this is doing it during translation to v0.
        """

        # -----------------------------------------------------------------------------
        # Common
        # -----------------------------------------------------------------------------
        if l.relation == b.common.range:
            # python/PyRel range has inclusive start but exclusive stop, whereas LQP/Rel has
            # inclusive stop, so we have to decrement the stop index by 1
            # tmp, lookup = self.decrement(l, 1, ctx)
            # return [tmp] + self.cast(lookup, [v0_types.Int64, v0_types.Int64, v0_types.Int64, v0_types.Int64], ctx)
            return self.decrement(l, 1, ctx)

        elif l.relation == b.core.cast:
            return v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.TASK),
                relation=v0_builtins.builtin_relations_by_name["cast"], # type: ignore
                args=(
                    self.translate_value(l.args[0], l, Context.MODEL),
                    self.translate_value(l.args[1], l, Context.TASK),
                    self.translate_value(l.args[2], l, Context.TASK)
                ),
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )

        # -----------------------------------------------------------------------------
        # Strings
        # -----------------------------------------------------------------------------

        elif l.relation == b.strings.len:
            lookup = v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                relation=v0_builtins.builtin_relations_by_name["num_chars"], # type: ignore
                args=(
                    self.translate_value(l.args[0], l, ctx),
                    self.translate_value(l.args[1], l, ctx),
                ),
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )
            return self.cast(lookup, ctx)

        elif l.relation == b.strings.split:
            # swap first two args because that's how rel and v0 expect them
            args = l.args[1], l.args[0], l.args[2], l.args[3]
            return self.adjust_index(l, [2], ctx, args=args)

        elif l.relation == b.strings.substring:
            return self.adjust_index(l, [1], ctx)

        # -----------------------------------------------------------------------------
        # RE
        # -----------------------------------------------------------------------------
        elif l.relation == b.re.regex_match_all:
            return self.adjust_index(l, [2], ctx)

        # -----------------------------------------------------------------------------
        # Math
        # -----------------------------------------------------------------------------
        elif l.relation == b.math.degrees:
            # degrees(radians, x) = /(radians, (pi/180.0), x)
            return v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                relation=v0_builtins.builtin_relations_by_name["/"], # type: ignore
                args=(
                    self.translate_value(l.args[0], l, ctx),
                    v0.Literal(v0_types.Float, pi / 180.0),
                    self.translate_value(l.args[1], l, ctx)
                ),
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )
        elif l.relation == b.math.radians:
            # radians(degrees, x) = /(degrees, (180.0/pi), x)
            return v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                relation=v0_builtins.builtin_relations_by_name["/"], # type: ignore
                args=(
                    self.translate_value(l.args[0], l, ctx),
                    v0.Literal(v0_types.Float, 180.0 / pi),
                    self.translate_value(l.args[1], l, ctx)
                ),
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )

        # -----------------------------------------------------------------------------
        # Datetime
        # -----------------------------------------------------------------------------
        #
        # elif l.relation == b.datetime.date_range:
        #     return self.rewrite_date_range(l, parent, ctx)
        # elif l.relation == b.datetime.datetime_range:
        #     pass

        elif l.relation == b.datetime.date_add:
            return v0.Lookup(
                engine=self.translate_reasoner(l.reasoner, l, Context.MODEL),
                relation=self.translate_relation(l.relation, l, ctx), # type: ignore
                args=self.translate_seq(l.args, l, ctx), # type: ignore
                annotations=self.translate_frozen(l.annotations, l, ctx) # type: ignore
            )

        if b.datetime[l.relation.name] is l.relation:
            return self.adjust_index(l, [], ctx, no_outputs=True)

        v0_rel = self.translate_node(l.relation, l, ctx)
        if isinstance(v0_rel, v0.Relation) and any(f.type == v0_types.Int64 for f in v0_rel.fields):
            return self.adjust_index(l, [], ctx)

        return None


    # TODO - Currently implemented in the std.datetime itself.
    # def rewrite_date_range(self, l: mm.Lookup, parent, ctx: Context):
    #     start, end, freq, date = l.args
    #     result = []

    #     range_end = num_days = mm.Var(type=b.core.Number, name="num_days")
    #     result.append(mm.Lookup(b.datetime.dates_period_days, (start, end, num_days)))

    #     assert isinstance(freq, mm.Literal) and isinstance(freq.value, str)
    #     if freq.value in ["W", "M", "Y"]:
    #         range_end = mm.Var(type=b.core.Number, name="range_end")
    #         x = mm.Var(type=b.core.Number, name="x")
    #         result.append(mm.Lookup(b.core.mul, (num_days, mm.Literal(b.core.Float, _days[freq.value]), x)))
    #         result.append(mm.Lookup(b.math.ceil, (x, range_end)))

    #     # date_range is inclusive. add 1 since std.range is exclusive
    #     tmp = mm.Var(type=b.core.Number, name="tmp")
    #     result.append(mm.Lookup(b.core["+"], (range_end, mm.Literal(b.core.Number, 1), tmp)))
    #     ix = mm.Var(type=b.core.Number, name="ix")
    #     result.append(mm.Lookup(b.common.range, (mm.Literal(b.core.Number, 0), tmp, mm.Literal(b.core.Number, 1), ix)))

    #     tmp2 = mm.Var(type=b.core.Number, name="tmp2")
    #     result.append(mm.Lookup(_periods[freq.value], (ix, tmp2)))
    #     result.append(mm.Lookup(b.datetime.date_add, (start, tmp2, date)))

    #     result.append(mm.Lookup(b.core[">="], (end, date)))

    #     return self.translate_node(mm.Logical(
    #         body=tuple(result)
    #     ))



from __future__ import annotations

from v0.relationalai.semantics.internal import internal as b
from .std import _make_expr

def parse_float(value: b.Producer|str) -> b.Expression:
    return _make_expr("parse_float", value, b.Float.ref("res"))
